/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * clntserv/clntserv.c   Implementation of client/server communications.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This file implements a small library for client/server communications
 * over TCP/IP and Unix domain sockets.
 * 
 * $Id: clntserv.c,v 1.6 2000/04/28 09:33:14 neunhoef Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>

#include "clntserv.h"


/* Configuration of behaviour: */


/* If this variable is non-zero, most errors lead to an exit(1) call: */
/* The default is 1. */
int cs_exit_on_error = 1;

/* If this variable is non-zero, the library automatically tries to do some 
   input/output via `cs_do_io' whenever a non-blocking call is issued.
   The default is 1. */
int cs_automatic_io = 1;

/* The following value in seconds is the timeout for blocking I/O. Set to
   zero if you do not want any timeout. The default value is the 
   abovementioned constant CS_BLOCKING_TIMEOUT. */
int cs_blocking_timeout = CS_BLOCKING_TIMEOUT;


/* Global data for the connections to the servers. */

typedef struct cs_server {
  int fd;                /* file descriptor, -1 means not used */
  CS_REQTYPE *current;   /* current request, sending or receiving */
} CS_SERVER;


/* Global data in the client: */

/* This list is updated by the library functions, but is present in the
   client. A fd==-1 means that the entry was never used and fd==-2 means, 
   that it was freed by a closed connection. A positive number is a number 
   of a file descriptor. */
static CS_SERVER cs_serverlist[CS_MAXSERVERS];

/* This flag is 0 until the first call to cs_connect: */
static int cs_client_initialized = 0;

/* Beginning of doubly linked list of req.: */
CS_REQTYPE *cs_req_first = NULL;
/* End of doubly linked list of requests: */
CS_REQTYPE *cs_req_last = NULL;    
/* Total number of bytes in question buffers of requests: */
size_t cs_question_bytes = 0;
/* Total number of bytes in answer buffers of requests: */
size_t cs_answer_bytes = 0;
/* Total number of requests in queue: */
long cs_number_requests = 0;
/* Total number of answered (successful or with error!) requests in queue: */
long cs_number_answered = 0;


/* Some useful things: */

ssize_t readbyselect(int fd, void *buf, size_t count, int skipselect)
/* This function implements non-blocking input by using `select': It tries
   to read up to `count' bytes from the file descriptor `fd' into the
   buffer pointed to by `buf'. It returns the number of bytes actually
   read (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). It first calls `select'
   with a timeout of zero and tries to read count bytes. This is
   repeated until either there is no more data present, the connection
   is broken or `count' bytes have been read. If the flag `skipselect'
   is !=0, we skip the first select assuming that `select' was called
   right before the call to this function. */
{
  char *buffer;         /* Position to read to */
  fd_set rfds;          /* for `select' */
  int ret;              /* Return value of `select' */
  struct timeval t;     /* For timeout */
  ssize_t retread;      /* Return value of `read' */

  buffer = (char *) buf;
  while (count > 0) {
    if (skipselect) {
      ret = 1;   /* We assume a select just before the call to us */
      skipselect = 0;   /* do not skip subsequent `select's */
    } else {
      FD_ZERO(&rfds);
      FD_SET(fd,&rfds);
      t.tv_sec = 0;    /* return immediately */
      t.tv_usec = 0;
      ret = select(fd+1,&rfds,NULL,NULL,&t);
      if (ret < 0)    /* This is an error! */
	return -1;
    }
    if (ret > 0) {   /* Our file descriptor is ready to read! */
      /* now we can read: */
      retread = read(fd,buffer,count);
      if (retread <= 0)   /* 0 means the connection is broken! */
	return -1;

      /* We got at least something: */
      count -= retread;
      buffer += retread;
    } else
      return (buffer - (char *) buf);  /* No more data there */
  }
  return (buffer - (char *) buf);  /* No more data there */
}

ssize_t writebyselect(int fd, void *buf, size_t count, int skipselect)
/* This function implements non-blocking output by using `select': It tries
   to write up to `count' bytes to the file descriptor `fd' from the
   buffer pointed to by `buf'. It returns the number of bytes actually
   written (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). It first calls `select'
   with a timeout of zero and tries to write count bytes. This is
   repeated until either the file descriptor cannot take more data,
   the connection is broken or `count' bytes have been written. If the
   flag `skipselect' is !=0, we skip the first select assuming that
   `select' was called right before the call to this function. */
{
  char *buffer;         /* Position to write from */
  fd_set wfds;          /* for `select' */
  int ret;              /* Return value of `select' */
  struct timeval t;     /* For timeout */
  ssize_t retwrite;     /* Return value of `write' */

  buffer = (char *) buf;
  while (count > 0) {
    if (skipselect) {
      ret = 1;   /* We assume a select just before the call to us */
      skipselect = 0;   /* do not skip subsequent `select's */
    } else {
      FD_ZERO(&wfds);
      FD_SET(fd,&wfds);
      t.tv_sec = 0;    /* return immediately */
      t.tv_usec = 0;
      ret = select(fd+1,NULL,&wfds,NULL,&t);
      if (ret < 0)    /* This is an error! */
	return -1;
    }
    if (ret > 0) {   /* Our file descriptor is ready to read! */
      /* now we can write: */
      retwrite = write(fd,buffer,count);
      if (retwrite <= 0)   /* 0 means the connection is broken! */
	return -1;

      /* We wrote at least something: */
      count -= retwrite;
      buffer += retwrite;
    } else
      return (buffer - (char *) buf);  /* No more possibility to write */
  }
  return (buffer - (char *) buf);  /* No more possibility to write */
}


ssize_t readblock(int fd, void *buf, size_t count)
/* This function implements blocking input by using `select': It tries
   to read up to `count' bytes from the file descriptor `fd' into the
   buffer pointed to by `buf'. It returns the number of bytes actually
   read (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). If necessary, multiple `read's are
   done but if a timeout occurs, the function returns. */
{
  char *buffer;         /* Position to read to */
  fd_set rfds;          /* for `select' */
  int ret;              /* Return value of `select' */
  struct timeval t;     /* For timeout */
  ssize_t retread;      /* Return value of `read' */

  buffer = (char *) buf;
  while (count > 0) {
    FD_ZERO(&rfds);
    FD_SET(fd,&rfds);
    /*  // printf("readblock:");fflush(stdout); */
    if (cs_blocking_timeout) {
      t.tv_sec = cs_blocking_timeout;    /* return after some time */
      t.tv_usec = 0;
      ret = select(fd+1,&rfds,NULL,NULL,&t);
    } else
      ret = select(fd+1,&rfds,NULL,NULL,NULL);

    /*  // printf("%d\n",ret);fflush(stdout); */
    if (ret < 0)    /* This is an error! */
      return -1;
    if (ret > 0) {   /* Our file descriptor is ready to read! */
      /* now we can read: */
      retread = read(fd,buffer,count);
      if (retread <= 0)   /* 0 means the connection is broken! */
	return -1;

      /* We got at least something: */
      count -= retread;
      buffer += retread;
    } else
      return (buffer - (char *) buf);  /* No more data there */
  }
  return (buffer - (char *) buf);  /* No more data there */
}

ssize_t writeblock(int fd, void *buf, size_t count)
/* This function implements blocking output by using `select': It tries
   to write up to `count' bytes to the file descriptor `fd' from the
   buffer pointed to by `buf'. It returns the number of bytes actually
   written (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). If necessary, multiple `write's are
   done but if a timeout occurs, the function returns. */
{
  char *buffer;         /* Position to write from */
  fd_set wfds;          /* for `select' */
  int ret;              /* Return value of `select' */
  struct timeval t;     /* For timeout */
  ssize_t retwrite;      /* Return value of `write' */

  buffer = (char *) buf;
  while (count > 0) {
    FD_ZERO(&wfds);
    FD_SET(fd,&wfds);
    if (cs_blocking_timeout) {
      t.tv_sec = cs_blocking_timeout;    /* return after some time */
      t.tv_usec = 0;
      ret = select(fd+1,NULL,&wfds,NULL,&t);
    } else
      ret = select(fd+1,NULL,&wfds,NULL,NULL);

    if (ret < 0)    /* This is an error! */
      return -1;
    if (ret > 0) {   /* Our file descriptor is ready to write! */
      /* now we can write: */
      retwrite = write(fd,buffer,count);
      if (retwrite <= 0)   /* 0 means the connection is broken! */
	return -1;

      /* We got at least something: */
      count -= retwrite;
      buffer += retwrite;
    } else
      return (buffer - (char *) buf);  /* No more possibility to write */
  }
  return (buffer - (char *) buf);  /* Done */
}

CS_BUFTYPE cs_new_buffer(size_t startlen)
/* This function allocates a new buffer with length startlen and sets
   up the struct for this. It returns a struct with `buf' entry NULL,
   if the allocation does not work. If `cs_exit_on_error' is non-zero,
   the program is terminated instead. If `startlen' is zero, the struct
   returned has `buf' entry NULL which is not an error. This struct can
   be used to call non-blocking requests where the answer buffer is
   allocated automatically. */
{
  CS_BUFTYPE buf;
  
  /* The trivial case: */
  if (startlen == 0) {
    buf.buf = NULL;
    buf.buflen = 0;
    buf.len = 0;
    return buf;
  }

  /* real case: */
  buf.buf = malloc(startlen);
  if (buf.buf == NULL) { 
    if (cs_exit_on_error) exit(1);   /* we leave the program */
    buf.buflen = 0;
  } else
    buf.buflen = startlen;
  buf.len = 0;
  return buf;
}

int cs_ensure_bufferlength(CS_BUFTYPE *buf,size_t len)
/* This function checks, whether the buffer `buf' is at least `len'
   bytes long, if not, `realloc' is called. Returns 0 if this worked
   and -1 otherwise. */
{
  if (buf->buflen < len) {
    buf->buf = realloc(buf->buf,len);
    if (buf->buf == NULL) {
      buf->buflen = 0;
      return -1;
    }
    buf->buflen = len;
  }
  return 0;
}

/* For the client: */

long cs_connect(char *name,unsigned short int port)
/* Opens connection to a server. `name' is an internet hostname,
   `port' a TCP port number. The result is an identifier for this
   connection (non-negative number) if everything is OK and otherwise
   a negative number. If the global flag `cs_exit_on_error' is
   non-zero, this function does not return if the connection is not
   established, but calls `exit(1)'. */
{
  int i;    /* just a counter */
  struct hostent *serverhost;
  struct sockaddr_in sockaddr;   /* address of sockets with port */
  struct protoent *protocol;    /* protocol: tcp */

  if (!cs_client_initialized) {
    for (i = 0;i < CS_MAXSERVERS;i++)
      cs_serverlist[i].fd = -1;
    cs_client_initialized = 1;  /* never do this again */
  }

  for (i = 0;i < CS_MAXSERVERS && cs_serverlist[i].fd >= 0;i++);

  if (i >= CS_MAXSERVERS) {
    fprintf(stderr,"Maximal number of servers reached: %d\n",CS_MAXSERVERS);
    if (cs_exit_on_error) exit(1); else return -1;
  }

  /* Fetch protocol tcp: */
  protocol = getprotobyname("tcp");
  if (!protocol) {
    fprintf(stderr,"Protocol tcp unknown!\n");
    if (cs_exit_on_error) exit(1); else return -2;
  }

  serverhost = gethostbyname(name);
  if (serverhost == NULL) {
    fprintf(stderr,"Hostname not found!\n");
    if (cs_exit_on_error) exit(1); else return -3;
  }

  /* Fetch a socket: */
  cs_serverlist[i].fd = socket(AF_INET,SOCK_STREAM,protocol->p_proto);
  if (cs_serverlist[i].fd < 0) {
    cs_serverlist[i].fd = -2;  /* no longer used */
    perror("Could not fetch socket");
    if (cs_exit_on_error) exit(1); else return -4;
  }

  /* Connect: */
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons(port);
  memcpy(&(sockaddr.sin_addr.s_addr),serverhost->h_addr_list[0],
	 serverhost->h_length);
  if (connect(cs_serverlist[i].fd, (struct sockaddr *)(&sockaddr), 
      sizeof(sockaddr))) {
    close(cs_serverlist[i].fd);
    cs_serverlist[i].fd = -2;
    perror("Error in connect call");
    if (cs_exit_on_error) exit(1); else return -5;
  }

  cs_serverlist[i].current = NULL;

  return i;
}

long cs_request(long server, long magic, 
		CS_BUFTYPE question, CS_BUFTYPE *answer)
/* This function sends a request and blocks until the answer has
   arrived. `server' is a server identifier as returned by
   `cs_connect', `magic' is a number which specifies the service
   required from the server. `question' specifies the input for the
   request. `*answer' specifies a buffer for the reply, which must not
   be longer than `answer->buflen'. `answer->len' is modified
   according to the length of the result. No memory allocation happens
   within this function. If everything is OK, 0 is returned and
   otherwise a negative number. If the global flag `cs_exit_on_error'
   is non-zero, this function does not return if the connection is not
   established, but calls `exit(1)'. Note that before blocking I/O can be
   done with the server, all pending non-blocking requests are made! */
{
  unsigned long buffer[2];
  ssize_t res;

  /* We check whether the server is ok: */
  if (server < 0 || server >= CS_MAXSERVERS) {
    fprintf(stderr,"Illegal server number!\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }

  if (cs_serverlist[server].fd < 0) {  /* No server with this number */
    fprintf(stderr,"No server with this number!\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }

  /* We wait until the server is free of non-blocking requests: */
  while (cs_serverlist[server].current) cs_do_io();
    
  buffer[0] = htonl(magic);  buffer[1] = htonl(question.len);
  res = writeblock(cs_serverlist[server].fd,buffer,2 * sizeof(unsigned long));
  if (res < (ssize_t) (2 * sizeof(unsigned long)) ) {
    fprintf(stderr,"Could not write magic and len.\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }
  res = writeblock(cs_serverlist[server].fd,question.buf,question.len);
  if (res < (ssize_t) question.len) {
    fprintf(stderr,"Could not write full request.\n");
    if (cs_exit_on_error) exit(1); else return -2;
  }
  res = readblock(cs_serverlist[server].fd,buffer,sizeof(unsigned long));
  if (res < (ssize_t) sizeof(unsigned long)) {
    fprintf(stderr, "Could not read answer length (connection broken?).\n");
    if (cs_exit_on_error) exit(1); else return -3;
  }
  buffer[0] = ntohl(buffer[0]);
  if (answer->buflen < buffer[0]) {
    fprintf(stderr,"Answer buffer too small!\n");
    if (cs_exit_on_error) exit(1); else return -4;
  }
  res = readblock(cs_serverlist[server].fd,answer->buf,buffer[0]);
  if (res < (ssize_t) buffer[0]) {
    fprintf(stderr,"Could not read answer.\n");
    if (cs_exit_on_error) exit(1); else return -5;
  }
  answer->len = buffer[0];
  return 0;
}


long cs_request_malloc(long server, long magic, CS_BUFTYPE question,
		       CS_BUFTYPE *answer)
/* This function sends a request and blocks until the answer has
   arrived. `server' is a server identifier as returned by
   `cs_connect', `magic' is a number which specifies the service
   required from the server. `question' specifies the input for the
   request. `*answer' specifies a buffer for the reply. If
   `answer->buf' is NULL, a new buffer of the correct length is
   malloced. Otherwise, `*answer' specifies an already existing
   buffer, which must have been allocated by `malloc'. The function is
   allowed to issue `realloc', if the buffer is not long enough for
   the answer. In this case `answer->buflen' is modified
   accordingly. `realloc' is not called if the buffer is too long for
   the answer. However `answer->len' is modified according to the
   length of the result. The result is 0 if everything was OK and a
   negative number if anything was wrong.  If the global flag
   `cs_exit_on_error' is non-zero, this function does not return if
   the connection is not established, but calls `exit(1)'. Note that
   it is the responsibility of the caller to free the answer buffer
   once it is no longer needed. */
{
  unsigned long buffer[2];
  long res;

  /* We check whether the server is ok: */
  if (server < 0 || server >= CS_MAXSERVERS) {
    fprintf(stderr,"Illegal server number!\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }

  if (cs_serverlist[server].fd < 0) {  /* No server with this number */
    fprintf(stderr,"No server with this number!\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }

  /* We wait until the server is free of non-blocking requests: */
  while (cs_serverlist[server].current) cs_do_io();
    
  buffer[0] = htonl(magic);  buffer[1] = htonl(question.len);
  res = writeblock(cs_serverlist[server].fd,buffer,2 * sizeof(unsigned long));
  if (res < (ssize_t) (2 * sizeof(unsigned long)) ) {
    fprintf(stderr,"Could not write magic and len.\n");
    if (cs_exit_on_error) exit(1); else return -1;
  }
  res = writeblock(cs_serverlist[server].fd,question.buf,question.len);
  if (res < (ssize_t) question.len) {
    fprintf(stderr,"Could not write full request.\n");
    if (cs_exit_on_error) exit(1); else return -2;
  }
  res = readblock(cs_serverlist[server].fd,buffer,sizeof(unsigned long));
  if (res < (ssize_t) sizeof(unsigned long)) {
    fprintf(stderr,"Could not read len (connection broken?).\n");
    if (cs_exit_on_error) exit(1); else return -3;
  }
  buffer[0] = ntohl(buffer[0]);
  if (answer->buf == NULL) {
    /*  // printf("malloc %ld bytes\n",buffer[0]);fflush(stdout); */
    answer->buf = malloc(buffer[0]);
    if (answer->buf == NULL) {
      fprintf(stderr,"Could not allocate answer buffer (blocking request).\n");
      if (cs_exit_on_error) exit(1); else return -4;
    }
    answer->buflen = buffer[0];
  } else {
    if (answer->buflen < buffer[0]) {
      answer->buf = realloc(answer->buf,buffer[0]);
      if (answer->buf == NULL) {
	fprintf(stderr,"Answer buffer too small and could not enlarge.\n");
	if (cs_exit_on_error) exit(1); else return -5;
      }
      answer->buflen = buffer[0];
    }
  }
  res = readblock(cs_serverlist[server].fd,answer->buf,buffer[0]);
  if (res < (ssize_t) buffer[0]) {
    fprintf(stderr,"Could not read answer.\n");
    if (cs_exit_on_error) exit(1); else return -6;
  }
  answer->len = buffer[0];
  return 0;
}


CS_REQTYPE *cs_request_internal(long server, long magic, CS_BUFTYPE question,
				CS_BUFTYPE answer,void *private,
				int shouldalloc)
/* Is used internally by `cs_request_nb' and `cs_request_nb_malloc' and
   does all the work. */
{
  CS_REQTYPE *req;   /* Pointer to new request. */

  if (server < 0 || server >= CS_MAXSERVERS) {
    fprintf(stderr,"Illegal server number!\n");
    if (cs_exit_on_error) exit(1); else return NULL;
  }

  if (cs_serverlist[server].fd < 0) {  /* No server with this number */
    fprintf(stderr,"No server with this number!\n");
    if (cs_exit_on_error) exit(1); else return NULL;
  }

  req = (CS_REQTYPE *) malloc(sizeof(CS_REQTYPE));
  if (req == NULL) {
    perror("Could not allocate request struct");
    if (cs_exit_on_error) exit(1); else return NULL;
  }

  /* Put this new request at end of queue: */
  req->prev = cs_req_last;
  req->next = NULL;
  if (cs_req_last)   /* The queue was not empty! */
    cs_req_last->next = req;
  else { /* The queue was empty! */
    cs_req_first = req;
  }
  cs_req_last = req;

  /* Now copy the buffers and initialize data: */
  req->question = question;
  req->alreadysent = 0;       /* first this refers to the header! */
  req->answer = answer;
  req->alreadyrecv = 0;
  if (shouldalloc) 
    req->allocateanswer = 1;  /* we will (re-)allocate the answer buffer */
  else
    req->allocateanswer = 0;  /* we won't touch the answer buffer */
  req->status = 0;
  req->server = server;   /* this is the server */
  req->header[0] = htonl(magic);         /* The magic and the length of the */
  req->header[1] = htonl(question.len); /* question make up the header */
  req->cancelled = 0;     /* this request was not yet cancelled */
  req->stopped = 0;       /* this request was not yet stopped */
  req->private = private; /* we store this along with our data such that
			     the calling program can use our queue. */
  /* Is there already a request for this server pending? */
  if (cs_serverlist[server].current == NULL)
    cs_serverlist[server].current = req;  /* no, here is one! */

  /* Now we update the global accounting: */
  cs_question_bytes += req->question.buflen;
  cs_number_requests++;

  /* We count the buffer length: */
  if (req->allocateanswer == 0)
    cs_answer_bytes += req->answer.buflen;

  /* All done, note, input/output is only done automatically, if the switch
     `cs_automatic_io' on! */

  if (cs_automatic_io)
    cs_do_io();

  return req;
}

CS_REQTYPE *cs_request_nb(long server, long magic, CS_BUFTYPE question,
		          CS_BUFTYPE answer,void *private)
/* This function handles a request. It is however a non-blocking
   version. Use `cs_isready' and `cs_whichready' to ask for results.
   `server' is a server identifier as returned by `cs_connect',
   `magic' is a number which specifies the service required from the
   server. `question' specifies the input for the request. `answer'
   specifies a buffer for the reply, which must not be longer than
   `answer.buflen'. `answer.len' is modified according to the length
   of the result. Note that this will only be modified, when the answer
   header is there, which occurs in some call to `cs_isready', `cs_do_io' or
   `cs_whichready'. The modification takes place in the instance of
   CS_BUFTYPE which is stored in the request queue. Do not expect that 
   your original struct is changed! No memory allocation happens within this
   function. The result is a pointer to the queueing data for this
   request. If the global switch `cs_automatic_io' is on, input/output is
   tried after queueing up the request by calling `cs_do_io'
   automatically. `private' is a void pointer that is stored with the
   request and used for nothing. */
{
  return cs_request_internal(server,magic,question,answer,private,0);
}

CS_REQTYPE *cs_request_nb_malloc(long server, long magic, 
  		 CS_BUFTYPE question,CS_BUFTYPE answer,void *private)
/* This function handles a request. It is however a non-blocking version.
   Use `cs_isready' and `cs_whichready' to ask for results.  `server'
   is a server identifier as returned by `cs_connect', `magic' is a
   number which specifies the service required from the server.
   `question' specifies the input for the request.  `answer' specifies
   a buffer for the reply.  If `answer.buf' is NULL, a new buffer of
   the correct length is malloced.  In this case, the corresponding
   `cs_forget' call will free this buffer again.  Otherwise, `answer'
   specifies an already existing buffer, which must have been
   allocated by `malloc'. The function is allowed to issue `realloc',
   if the buffer is not long enough for the answer.  In this case
   `answer.buflen' is modified accordingly.  `realloc' is not called
   if the buffer is too long for the answer.  However `answer.len' is
   modified according to the length of the result.  Note that this
   will only be modified, when the answer header is there, which
   occurs in some call to `cs_isready', `cs_do_io' or
   `cs_whichready'. The modification takes place in the instance of
   CS_BUFTYPE which is stored in the request queue. Do not expect that
   your original struct is changed! `cs_forget' will not free the
   buffer in the case where the original allocation was not done in
   this library.  It is the responsibility of the caller to free the
   answer buffer once it is no longer needed.  If the global switch
   `cs_automatic_io' is on, input/output is tried after queueing up
   the request by calling cs_do_io automatically. `private' is a void
   pointer that is stored with the request and used for nothing. */
{
  return cs_request_internal(server,magic,question,answer,private,1);
}

static void connection_lost(long server) 
/* This internal function is called, when something is wrong with the
   server `server'. This can be a lost connection or some memory allocation 
   problems clobbering the connection. The connection is closed and
   forgotten. All requests for this server are cancelled with error
   condition. If a request for the server `server' is already
   cancelled, it is deleted and buffers are freed. */
{
  CS_REQTYPE *reqp;      /* to run through the requests */
  CS_REQTYPE *reqpnext;  /* the next one */

  close(cs_serverlist[server].fd);
  cs_serverlist[server].fd = -2; 

  for (reqp = cs_serverlist[server].current;reqp;reqp = reqpnext) {
    reqpnext = reqp->next;
    /* All requests after the current one, which belong to this server: */
    if (reqp->server == server) {
      if (reqp->cancelled)
	cs_forget(reqp,1);   /* We try to get the memory back! */
      else {
	reqp->status = -1;  /* error condition! */
	cs_number_answered++;
      }
    }
  }
  fprintf(stderr,"Connection to server lost.\n");
  if (cs_exit_on_error) exit(1);
}

static CS_REQTYPE *next_request(long server)
/* This function is internal. It searches all requests for the first one
   that is for server `server' and that is not yet started and that is not
   stopped. */
{
  CS_REQTYPE *req;

  for (req = cs_req_first;req;req = req->next) {
    if (req->server == server &&    /* Yes! */
	req->status == 0 &&         /* not yet started or done */
	req->alreadysent == 0 &&    /* this would not be necessary? */
	req->stopped == 0)          /* not stopped */
      return req;
  }
  return NULL;
}

void cs_do_io(void)
/* This function does input and output on all open connections handled 
   by this library in a non-blocking way (using select). This can lead
   to the completion of requests, which means, that the answer is
   fully there. It is the responsibility of the user of this library
   to call this often enough, whatever that means... */
{
  fd_set rfds;               /* bit mask for reading on file descriptors */
  fd_set wfds;               /* bit mask for writing on file descriptors */
  int maxfd;                 /* largest number of file desc in use */
  int selres;                /* result of `select' call */
  ssize_t res;               /* Result of reads and writes */
  int skipselect;            /* Flag, whether we skip a select */
  CS_REQTYPE *reqp;          /* Pointer to a request */
  int fd;                    /* File descriptor */
  long i;                    /* Counter for the servers */
  struct timeval t;          /* for zero timeout */

  /* prepare for "select": */
  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  maxfd = -1;
  for (i = 0;i < CS_MAXSERVERS && cs_serverlist[i].fd != -1;i++) {  
    /* all connections */
    if (cs_serverlist[i].fd >= 0 && 
        cs_serverlist[i].current) {  /* there is a request pending */
      fd = cs_serverlist[i].fd;
      reqp = cs_serverlist[i].current;
      if (reqp->status == 0 || reqp->status == 1) /* we are writing */
	FD_SET(fd,&wfds);
      else   /* we are reading */
	FD_SET(fd,&rfds);

      if (fd > maxfd) maxfd = fd;
    }
  }

  /* Do we have to do anything? */
  if (maxfd == -1) return;

  t.tv_sec = 0;    /* return immediately */
  t.tv_usec = 0;
  selres = select(maxfd+1, &rfds, &wfds, NULL, &t);
  
  if (selres > 0) {  /* something on the sockets? */
    /* check open connections for requests or data to be sent */
    for (i = 0;i < CS_MAXSERVERS && cs_serverlist[i].fd != -1;i++) {    
      /* all connections */
      if (cs_serverlist[i].fd >= 0 &&
          cs_serverlist[i].current) { /* there is a request pending */
	reqp = cs_serverlist[i].current;
	fd = cs_serverlist[i].fd;

	if (reqp->status == 0 || reqp->status == 1) {  
	  /* we are sending the question: */
	  if (FD_ISSET(fd,&wfds)) {
	    skipselect = 1;
	    if (reqp->status == 0) {  /* we are sending the header */
	      /* we send the length back: */
	      res = write(fd, ((char *) reqp->header)+reqp->alreadysent,
			  2 * sizeof(unsigned long) - reqp->alreadysent);
	      skipselect = 0;
	      /* note that we use `write' instead of `writebyselect'
		 because we expect the header to be sent en block */
	      if (res <= 0) {
		fprintf(stderr, "Could not write header.\n");
		connection_lost(i);  /* cancel all requests */
		break;  /* next server */
	      }
	      reqp->alreadysent += res;
	      if (reqp->alreadysent == 2 * sizeof(unsigned long)) {
		reqp->status = 1;  /* now we send the data of the question */
		reqp->alreadysent = 0;
	      }
	    }  /* no else because after status==0 it could be changed to 1 */
	    if (reqp->status == 1) {  /* we are sending the question */
	      res = writebyselect(fd, 
		  (char *)(reqp->question.buf) + reqp->alreadysent, 
		  reqp->question.len - reqp->alreadysent,
		  skipselect);
	      if (res < 0) {
		fprintf(stderr, "Could not write part of question.\n");
		connection_lost(i);  /* cancel all requests */
		break;
	      }
	      reqp->alreadysent += res;
	      if (reqp->alreadysent == reqp->question.len) {
		/* The transfer is ready! */
		reqp->status = 2;
		reqp->alreadyrecv = 0;
	      }
	    }
	  }
	} else {   /* reqp->status > 1   ==> we are reading */
	  if (FD_ISSET(fd,&rfds)) {
	    skipselect = 1;
	    if (reqp->status == 2) {  /* we are waiting for the header */
	      res = read(fd, ((char *) (&(reqp->answerlength))) + 
			 reqp->alreadyrecv, 
			 sizeof(unsigned long) - reqp->alreadyrecv);
	      skipselect = 0;
	      /* note that we use `read' instead of `readbyselect'
		 because we expect the header to arrive en block */
	      if (res <= 0) {
		fprintf(stderr, "Could not read length of answer.\n");
		connection_lost(i);   /* cancel all requests */
		break;
	      }
	      reqp->alreadyrecv += res;
	      if (reqp->alreadyrecv == sizeof(unsigned long)) {
		reqp->status = 3;
		reqp->answer.len = ntohl(reqp->answerlength);
		reqp->alreadyrecv = 0;
		/* Now check the size of the answer buffer: */
		if (reqp->allocateanswer) { /* we have to allocate memory */
		  if (reqp->answer.buf == NULL) { /* nothing there yet */
		    /* // printf("malloc %ld bytes\n",(long)reqp->answer.len);fflush(stdout); */
		    reqp->answer.buf = malloc(reqp->answer.len);
		    /* bookkeeping; */
		    cs_answer_bytes += reqp->answer.len;
		    reqp->allocateanswer = 2;  /* we free it later on */
		    if (reqp->answer.buf == NULL) {
		      perror("Could not allocate answer buffer");
		      if (cs_exit_on_error) exit(1); 
		      else {
			connection_lost(i);  /* cancel all requests */
			break;
		      }
		    }
		    reqp->answer.buflen = reqp->answer.len;
		  } else {  /* Now check the buffer length: */
		    if (reqp->answer.len > reqp->answer.buflen) {
		      /* Do the bookkeeping: */
		      cs_answer_bytes -= reqp->answer.buflen;
		      reqp->answer.buf = realloc(reqp->answer.buf,
						  reqp->answer.len);
		      if (reqp->answer.buf == NULL) {
			perror("Cannot reallocate answer buffer");
			connection_lost(i);  /* cancel all requests */
			break;
		      }
		      reqp->answer.buflen = reqp->answer.len;
		      cs_answer_bytes += reqp->answer.buflen;
		    }
		  }
		} else {   /* we only check: */
		  if (reqp->answer.len > reqp->answer.buflen) {
		    fprintf(stderr,"Answer buffer too small.\n");
		    connection_lost(i);  /* cancel all requests */
		    break;
		  }
		}
	      }
	    }  /* no else because after status==2 it could be changed to 3 */
	    if (reqp->status == 3) {  /* we are receiving the answer */
	      res = readbyselect(fd, 
		 (char *) (reqp->answer.buf) + reqp->alreadyrecv, 
		 reqp->answer.len - reqp->alreadyrecv,
		 skipselect);
	      if (res < 0) {
		fprintf(stderr, "Could not read part of answer.\n");
		connection_lost(i);   /* cancel all requests */
		break;
	      }
	      reqp->alreadyrecv += res;
	      if (reqp->alreadyrecv == reqp->answer.len) {
		/* This request is completed: */
		reqp->status = 4;
		cs_number_answered++;
		/* Is there another request (later in the queue) for the
		   same server with number `i'? */
		cs_serverlist[i].current = next_request(i);

		/* Was it cancelled? */
		if (reqp->cancelled)
		  cs_forget(reqp,1);  /* this deletes this request */
	      }
	    }
	  }
	}
      }
    }
  }
}

int cs_isready(CS_REQTYPE *request)
/* This function checks, whether the request specified by `request' is
   completed, which means, that the answer is fully there. Note that input
   or output is done before that, only if `cs_automatic_io' is non-zero, so
   otherwise you should probably call `cs_do_io' beforehand. It is
   guaranteed that `cs_isready' needs nearly no CPU time. The return value
   is 0, if the request is not yet completed, 1, if it is sucessfully
   completed and negative, if an error occurred. The answer can be found in
   the buffer, which is specified by the exact instance of `CS_BUFTYPE' the
   request was started with. */
{
  /* Do we have to do input/output beforehand? */
  if (cs_automatic_io) 
    cs_do_io();

  if (request->status == 4)
    return 1;
  else if (request->status == -1)
    return -1;
  else
    return 0;
}

CS_REQTYPE *cs_whichready(long server)
/* This function checks, whether there is a request which was sent to
   the server `server', which is already completed. If so, the one
   which was sent first (out of those completed) ist returned. Note
   that no "overtaking" can occur for the requests sent to one
   specific server. If `server' is equal to -1, all requests to all
   servers are considered. Note that in this case "overtaking" can
   occur, it may very well be that one server answers more
   quickly. The function returns the request which was sent first (out
   of those already completed). Note that input or output is done
   before that, only if `cs_automatic_io' is non-zero, so otherwise you
   should probably call `cs_do_io' beforehand. It 
   is guaranteed that `cs_whichready' needs nearly no CPU time. The
   return value is NULL, if no requests are completed. */
{
  CS_REQTYPE *reqp;

  /* Do we have to do input/output beforehand? */
  if (cs_automatic_io) 
    cs_do_io();
  
  for (reqp = cs_req_first;reqp;reqp = reqp->next) {
    if ((server == -1 || reqp->server == server) &&
	cs_isready(reqp))
      return reqp;
  }
  return NULL;
}

void cs_forget(CS_REQTYPE *request, int freebuffers)
/* This function is used to clear a completed request out of the data
   structures. The request `request' must be completed and is deleted
   in the doubly linked list. If `freebuffers' is equal to 0, the
   answer buffer is only freed, if this library has allocated it.
   If `freebuffers' is non-zero, the answer buffer and the question
   buffer are freed automatically. */
{
  /* Just to be sure: */
  if ((request->status == 0 && request->alreadysent != 0) ||
      (request->status >= 1 && request->status <= 3)) {
    /* This was illegal! We do nothing! */
    fprintf(stderr,"Tried to `cs_forget' request which is doing I/O!\n");
    return;
    /* Should we just mark it cancelled??? */
  }

  /* Do the bookkeeping: */
  cs_question_bytes -= request->question.buflen;
  cs_answer_bytes -= request->answer.buflen;
  cs_number_requests--;
  if (request->status == 4 || request->status == -1)
    cs_number_answered--;

  if (!freebuffers) {
    /* Did we ourself allocate the answer buffer? */
    if (request->allocateanswer == 2) {
      free(request->answer.buf);
      request->answer.buf = NULL;
      request->answer.buflen = 0;
    }
  } else {
    /* We free all buffers: */
    free(request->answer.buf);
    free(request->question.buf);
  }

  /* Take it out of the queue: */
  if (request->prev)  /* this is not the first one */
    request->prev->next = request->next;
  else
    cs_req_first = request->next;
  if (request->next)  /* this is not the last one */
    request->next->prev = request->prev;
  else
    cs_req_last = request->prev;

  /* Free the memory: */
  free(request);
}

void cs_stop_nb(CS_REQTYPE *req)
/* Stops the request `req'. This means that its `stopped' flag is set. This 
   can be undone with `cs_continue_nb'. A stopped request for which I/O
   already is started will be completed regardless such that the protocol
   with the server is not broken. However, when the server looks for the
   next request for a server, stopped requests are not considered. For an
   already completed request it means nothing to be stopped. */
{
  req->stopped = 1;
}

void cs_stopall_nb(long server)
/* Same as `cs_stop_nb' for all requests going to the server `server' or to
   all servers, if `server' is -1. */
{
  CS_REQTYPE *r;

  for (r = cs_req_first;r;r = r->next)
    if (server == -1 || server == r->server)
      r->stopped = 1;
}

void cs_continue_nb(CS_REQTYPE *req)
/* Continues the request `req'. This means, that the `stopped' flag is
   unset and - if the corresponding server has no current request - the
   first request which is not yet finished and is not stopped is begun,
   which can be `req'. If `cs_automatic_io' is set, `cs_do_io' is called
   afterwards. */

{
  if (req->stopped) {   /* we do not have to do anything otherwise */
    req->stopped = 0;
    if (cs_serverlist[req->server].current == NULL)
      /* nothing going on, so activate a new request */
      cs_serverlist[req->server].current = next_request(req->server);
  }

  if (cs_automatic_io)
    cs_do_io();
}

void cs_continueall_nb(long server)
/* Same as `cs_continue_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */

{
  CS_REQTYPE *req;

  for (req = cs_req_first;req;req = req->next) {
    if (server == -1 || req->server == server) {
      if (req->stopped) {   /* we do not have to do anything otherwise */
	req->stopped = 0;
	if (cs_serverlist[req->server].current == NULL) {
	  /* nothing going on, so activate a new request */
	  if (req->status == 0)  /* not yet done */
	    cs_serverlist[req->server].current = req;
	    /* note that we do not need `next_request' because we ourself
	       ran through the queue from beginning to here! */
	}
      }
    }
  }

  if (cs_automatic_io)
    cs_do_io();
}

void cs_repeat_nb(CS_REQTYPE *req)
/* The request `req' is repeated. This means nothing if the request is not
   yet begun (no I/O yet). If it is already completed it will be marked as
   `not yet started' and will be sent again to the server. The previous
   answer will be overwritten. Nothing happens to a request which has
   already started to do I/O but is not yet finished. The user has to make
   sure, that this is handled! */

{
  if (req->status == 4 ||    /* Request ready without error */
      req->status == -1) {   /* Request ready with an error */
    req->status = 0;         /* Start it anew */
    req->alreadysent = 0;
    if (cs_serverlist[req->server].current == NULL)
      /* nothing going on, so activate a new request: */
      cs_serverlist[req->server].current = next_request(req->server);
  }

  if (cs_automatic_io)
    cs_do_io();
}
    
void cs_repeatall_nb(long server)
/* Same as `cs_repeat_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */

{
  CS_REQTYPE *req;

  for (req = cs_req_first;req;req = req->next) {
    if (server == -1 || req->server == server) {
      if (req->status == 4 ||    /* Request ready without error */
	  req->status == -1) {   /* Request ready with an error */
	req->status = 0;         /* Start it anew */
	req->alreadysent = 0;
	if (cs_serverlist[req->server].current == NULL)
	  cs_serverlist[req->server].current = req;
	/* note that we do not need `next_request' because we ourself
	   ran through the queue from beginning to here! */
      }
    }
  }

  if (cs_automatic_io)
    cs_do_io();
}

void cs_cancel_nb(CS_REQTYPE *req)
/* The request `req' is cancelled. This has the following consequences: A
   request which has not yet started to do I/O is deleted by
   `cs_forget'. If I/O has already started, the flag `cancelled' is set,
   which means, that the request is deleted by `cs_forget' once it is
   completed, such that the protocol with the server is not disturbed. A
   request which is already completed is deleted by `cs_forget'. Note
   that the buffers for this request are freed, when the request is
   deleted! */
{
  if (req->status == 0 && req->alreadysent == 0)  /* not yet started I/O */
    cs_forget(req,1);
  else if (req->status == 4 || req->status == -1) /* already completed */
    cs_forget(req,1);
  else
    req->cancelled = 1;
}

void cs_cancelall_nb(long server)
/* Same as `cs_repeat_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */
{
  CS_REQTYPE *req;

  for (req = cs_req_first;req;req = req->next) {
    if (server == -1 || req->server == server) 
      cs_cancel_nb(req);
  }
}

/* For the server: */

typedef struct cs_clienttype {
  struct cs_clienttype *next;   /* linked list, last has NULL here */
  CS_BUFTYPE question;          /* The question, .buf==NULL means no
				   buffer there */
  size_t alreadyrecv;           /* Amount of data already sent */
  CS_BUFTYPE answer;            /* The output, .buf==NULL means no
				   answer */
  size_t alreadysent;           /* Amount of data already received */
  int status;                   /* 0 : no request there or header not fully
				   received, 1 : not fully recved, 2 : recved,
				   but header not fully sent, 3 : not fully
				   sent */
  int fd;                       /* socket file descriptor */
  long buffer[2];               /* magic number and length */
} CS_CLIENTTYPE;


CS_CLIENTTYPE *clients = NULL;     /* begin of linked list */

int cs_number_of_SIGPIPEs = 0;

void cs_sigpipe_handler(int i)

{
  i = i; /* NOTUSED */
  cs_number_of_SIGPIPEs++;
}

int cs_server_init_network(unsigned short int port)
/* Initialisiert die Netzwerkverbindung, also das "Lauschen" am Port
   port. Wenn etwas schief geht, wird ein Fehlercode ungleich 0 zurueck-
   gegeben, ansonsten der file descriptor (>= 0) fuer alles klar. */
{
  int value;                     /* for setsockopt */
  struct protoent *protocol;     /* protocol: tcp */
  struct sockaddr_in sockaddr;   /* address of socket and port */
  int sock;

  /* fetch protocol tcp: */  /* probably we could just assume 6 as number */
  protocol = getprotobyname("tcp");
  if (!protocol) {
    perror("Protocol tcp unknown");
    return -1;
  }

  /* fetch a socket: */
  sock = socket(AF_INET,SOCK_STREAM,protocol->p_proto);
  if (sock < 0) {
    perror("Could not get socket");
    return -2;
  }
  /* switch to "REUSEADDR": */
  value = 1;
  setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&value,sizeof(value));  

  /* connect socket with port: */
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons(port);
  /* all hosts are allowed to connect: */
  sockaddr.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock,(struct sockaddr *) &sockaddr,sizeof(sockaddr))) {
    perror("Could not bind");
    return -3;
  }

  /* accept connections: */
  if (listen(sock,CS_MAXQUEUE)) {
    perror("Cannot listen");
    return -4;
  }

  return sock;
}

void throw_out_client(CS_CLIENTTYPE *clp, CS_CLIENTTYPE *clpprev)
/* Only called internally, throws out client. */
{
  printf("Throwing out client!\n");
  close(clp->fd);
  if (clpprev == NULL)
    clients = clp->next;
  else
    clpprev->next = clp->next;
  free(clp->question.buf);
  free(clp->answer.buf);
  free(clp);
}

void cs_start_server(CS_WORKER workers[],unsigned short int port)
/* Starts the main loop of the server program. All data is
   initialized. This main loop handles all the connections, input and
   output. Once a complete request is received, the corresponding
   worker function is called, according to the magic value. This has
   to allocate the answer buffer and return the result in there. It
   does not have to worry about sending it. The buffer will be freed
   automatically, when all data is transferred. This function
   terminates never. */

{
  fd_set rfds;               /* bit mask for reading on file descriptors */
  fd_set wfds;               /* bit mask for writing on file descriptors */
  int maxfd;                 /* largest number of file desc in use */
  int selres;                /* result of `select' call */
  struct sockaddr_in newadr; /* incoming address */
  socklen_t addrlen;         /* length of address */
  int newclient;             /* file desc. for new client */
  char newadrstr[216];       /* ascii-Format of client address */
  long ret;                  /* return value of worker */
  CS_CLIENTTYPE *clp;        /* pointer to client data structure */
  CS_CLIENTTYPE *clpnext;    /* pointer to next clients data structure */
  CS_CLIENTTYPE *clpprev;    /* pointer to prev clients data structure */
  CS_CLIENTTYPE *newcl;      /* pointer to a new client */
  int sock;                  /* socket for incoming connections */
  int clianz;                /* number of clients */
  ssize_t res;               /* Result of reads and writes */
  CS_WORKER *worker;         /* For search in workers list */
  int skipselect;            /* Flag, whether we skip a select */
  struct sigaction sigact;   /* For the SIGPIPE signal handler */

  sock = cs_server_init_network(port);  /* this only returns, if everything
					   is OK */
  clianz = 0;   /* no connections yet */

  /* We catch the SIGPIPE signal: */
  sigact.sa_handler = SIG_IGN; /*    // cs_sigpipe_handler;  */
  sigemptyset(&(sigact.sa_mask));
  sigact.sa_flags = SA_RESETHAND;
  if (sigaction(SIGPIPE,&sigact,NULL) < 0) {  /* We catch SIGPIPE signal */
    perror("Cannot catch SIGPIPE signal");
    if (cs_exit_on_error) exit(1); else return;
  }

  for (;;) {   /* We exit only upon call from outside */
    /* prepare for "select": */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    maxfd = sock;
    for (clp = clients;clp;clp = clp->next) {
      if (clp->status == 0 || clp->status == 1) /* we are reading or waiting */
	FD_SET(clp->fd,&rfds);
      else   /* we are writing */
	FD_SET(clp->fd,&wfds);

      if (clp->fd > maxfd) maxfd = clp->fd;
    }
    FD_SET(sock,&rfds);

/*    printf("Number of SIGNALs: %d\n",cs_number_of_SIGPIPEs); */

    selres = select(maxfd+1, &rfds, &wfds, NULL, NULL);

    if (selres > 0) {  /* something on the sockets? */
      /* check open connections for requests or data to be sent */
      for (clp = clients,clpprev = NULL; clp ; clp = clpnext) {
	clpnext = clp->next;
	if (clp->status >= 2) {  /* we are sending the answer */
	  if (FD_ISSET(clp->fd,&wfds)) {
	    skipselect = 1;
	    if (clp->status == 2) {  /* we are sending the header */
	      /* we send the length back: */
	      res = write(clp->fd, ((char *) clp->buffer)+clp->alreadysent,
			  sizeof(unsigned long) - clp->alreadysent);
	      skipselect = 0;
	      /* note that we use `write' instead of `writebyselect'
		 because we expect the header to be sent en block */
	      if (res <= 0) {
		fprintf(stderr, "Could not write length of answer.\n");
		throw_out_client(clp,clpprev); 
		break;
	      }
	      clp->alreadysent += res;
	      if (clp->alreadysent == sizeof(unsigned long)) {
		clp->status = 3;
		clp->alreadysent = 0;
	      }
	    }  /* no else because after status==2 it could be changed to 3 */
	    if (clp->status == 3) {  /* we are sending the answer */
	      res = writebyselect(clp->fd, 
		  (char *)(clp->answer.buf) + clp->alreadysent, 
		  clp->answer.len - clp->alreadysent,
		  skipselect);
	      if (res < 0) {
		fprintf(stderr, "Could not write part of answer.\n");
		throw_out_client(clp,clpprev); 
		break;
	      }
	      clp->alreadysent += res;
	      if (clp->alreadysent == clp->answer.len) {
		/* The transfer is ready! */
		clp->status = 0;
		clp->alreadyrecv = 0;
	      }
	    }
	  }
	} else {/* clp->status < 2   ==> we are reading */
	  if (FD_ISSET(clp->fd,&rfds)) {
	    skipselect = 1;
	    if (clp->status == 0) {  /* we are waiting for a new request */
	      /* we ask for a magic number to distinguish the requests */
	      res = read(clp->fd, ((char *) clp->buffer) + clp->alreadyrecv, 
			 2*sizeof(unsigned long) - clp->alreadyrecv);
	      skipselect = 0;
	      /* note that we use `read' instead of `readbyselect'
		 because we expect the header to arrive en block */
	      if (res <= 0) {
		fprintf(stderr, "Could not read header.\n");
		throw_out_client(clp,clpprev); 
		break;
	      }
	      clp->alreadyrecv += res;
	      if (clp->alreadyrecv == 2*sizeof(unsigned long)) {
		clp->status = 1;
		clp->question.len = ntohl(clp->buffer[1]);
		clp->buffer[0] = ntohl(clp->buffer[0]);
		clp->alreadyrecv = 0;
		if (clp->question.len > clp->question.buflen) {
		  clp->question.buf = realloc(clp->question.buf,
					      clp->question.len);
		  if (clp->question.buf == NULL) {
		    fprintf(stderr, "Cannot allocate buffer for question.\n");
		    throw_out_client(clp,clpprev);
		    break;
		  }
		  clp->question.buflen = clp->question.len;
		}
	      }
	    }  /* no else because after status==0 it could be changed to 1 */
	    if (clp->status == 1) {  /* we are receiving a new request */
	      res = readbyselect(clp->fd, 
		 (char *) (clp->question.buf) + clp->alreadyrecv, 
		 clp->question.len - clp->alreadyrecv,
		 skipselect);
	      if (res < 0) {
		fprintf(stderr, "Could not read part of question.\n");
		throw_out_client(clp,clpprev); 
		break;
	      }
	      clp->alreadyrecv += res;
	      if (clp->alreadyrecv == clp->question.len) {
		for (worker = workers;
		     worker->magic && worker->magic != clp->buffer[0];
		     worker++);
		if (worker->magic == 0) {  /* Magic number not found */
		  fprintf(stderr,"Invalid magic number!\n");
		  throw_out_client(clp,clpprev); 
		  break;
		}
		ret = worker->worker(clp->question,&(clp->answer));
		/* note that the worker promises to handle the answer buffer
		   by calling realloc if necessary! */
		if (ret < 0) {
		  fprintf(stderr,"Worker routine returned an error!\n");
		  throw_out_client(clp,clpprev); 
		  break;
		}
		clp->status = 2;
		clp->alreadysent = 0;
		clp->buffer[0] = htonl(clp->answer.len);
	      }
	    }
	  }
	}
	clpprev = clp;
      }

      /* a new connection? */
      if (FD_ISSET(sock, &rfds) && clianz < CS_MAXCLIENTS) {
	addrlen = sizeof(newadr);
	newclient = accept(sock,(struct sockaddr *) &newadr,&addrlen);
	if (newclient < 0)
	  perror("error during accept");
	else {
	  strcpy(newadrstr,inet_ntoa(newadr.sin_addr)); 
	  printf("New connection from %s Port %d!\n",newadrstr,
		 ntohs(newadr.sin_port));
	  /* store new connection in our list: */
	  newcl = (CS_CLIENTTYPE *) malloc(sizeof(CS_CLIENTTYPE));
	  if (newcl == NULL) {
	    perror("Cannot allocate structure for new client");
	    close(newclient);
	  } else {
	    newcl->next = clients;
	    clients = newcl;
	    newcl->fd = newclient;
	    newcl->status = 0;
	    newcl->question.buf = NULL;
	    newcl->question.buflen = 0;
	    newcl->question.len = 0;
	    newcl->answer.buf = NULL;
	    newcl->answer.buflen = 0;
	    newcl->answer.len = 0;
	    newcl->alreadyrecv = 0;     /* no request begun to read */
	  }
	}
      }
    }
  }
  printf("Final number of signals: %d\n",cs_number_of_SIGPIPEs);
}

