/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * dcgeneric/dchashserv.c   Generic part of the hash server(s).
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This program is for storing and recognizing points of an orbit as a 
 * central service.
 * 
 *  $Id: dchashserv.c,v 1.6 2000/04/28 09:08:43 neunhoef Exp $
 *  
 */


/* The following changes the behaviour of custom.h : */
#undef DC_ISCLIENT

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "clntserv.h"

/* Our magic numbers for the identification of requests: */
#include "dcservices.h"

/* The utilities for hashes and lists of longs: */
#include "dcutil.h"


/* The things that have to be customized: */
#include "custom.h"


static HashType *hG;            /* main hash table */

static int fdvec, fdnrs;           /* file descriptors for hash table */

static LongList *KOrbRepLen;       /* lengths of representative lists */
static PtrList *KOrbRep;           /* representative lists of K-orbits */

LongList *hashservconf;        /* A list of numbers for configuration */
   /* the first one is the length of the hash, all further numbers can
      be used by the hash function and its initialization routine. */

long hGlen;   /* length of the hash, same as hashservconf->list[0] */
long dc_length_pt; /* length of a point in memory in bytes 
		      has to be set by the init function! */

long jobserver;   /* We need to talk to the jobserver */


int ReadKnownOrbits(void)
/* This function reads data from previous runs to go on with the
   calculation. It does not check for consistency! */
{
  long n, i, t[4];
  void *vv;

  lseek(fdnrs, 0, SEEK_SET);
  lseek(fdvec, 0, SEEK_SET);
  /* Every K-orbit has an information record in the "numbers" file
     which contains the number of the K-orbit, the number of points
     for representatives, the number of additional minimal points and
     the total number of points in the K-orbit (just for statistical
     purposes). */
  n = read(fdnrs, (char *) t, 4*sizeof(long));
  while (n == 4*sizeof(long)) {
    /* reading U-orbit representatives and put them in hash table and 
       orbits lists */
    printf("Reading data for K-orbit %ld from file.\n", t[0]); fflush(stdout);
    vv = malloc((t[1]+t[2]) * dc_length_pt);
    if (vv == NULL) {
      fprintf(stderr,"Cannot allocate memory for points!\n");
      exit(1);
    }
    if (read(fdvec, vv, (t[1]+t[2]) * dc_length_pt ) < 
	(t[1]+t[2]) * (signed long) dc_length_pt) {
      perror("Read error:");
      exit(1);
    }
    AppendZerosLongList(KOrbRepLen,t[0]+1);
    AppendZerosPtrList(KOrbRep,t[0]+1);
    KOrbRepLen->list[t[0]] = t[1];
    KOrbRep->list[t[0]] = vv;
    /* now add all points to hash table */
    for (i = 0; i < t[1] + t[2]; i++) {
      AddHT(hG, vv, dc_length_pt, t[0]);
      vv = (char *)vv + dc_length_pt;
    }
    n = read(fdnrs, (char *) t, 4*sizeof(long));
  }

  /* check that files don't contain garbage after this information: */
  if (n != 0) {
    fprintf(stderr,"End of .nrs file seems corrupted.\n");
    exit(7);
  }
  if (0 != read(fdvec, (char *) t, 1)) {
    fprintf(stderr,"End of .vec file seems corrupted.\n");
    exit(8);
  }

  return 0;
}

long hashquery(CS_BUFTYPE question,CS_BUFTYPE *answer)
/* This worker function handles a request of type HASHQUERY. This is a 
   request to lookup vectors in the hash table. The question buffer
   contains some points, each `dc_length_pt' bytes long. The length of the
   question `question.len' is divided by `dc_length_pt' to determine the
   number of points. The answer is a long for each input: either the
   number of the K-orbit this point lies in or HTFAIL, if not known. */
{
  long nr, i;
  void *v;
  long *r;

  nr = question.len / dc_length_pt;
  if (cs_ensure_bufferlength(answer,sizeof(long) * nr)) {
    fprintf(stderr,"Not enough memory for answer!\n");
    return -1;
  }
  answer->len = nr*sizeof(long);
  v = question.buf;
  r = answer->buf;
  for (i = 0; i < nr; i++) {
    *r++ = ValueHT(hG, v, dc_length_pt);
    v = (char *)v + dc_length_pt;
  }
  return 0;
}

long hashnew(CS_BUFTYPE question,CS_BUFTYPE *answer) 
/* This worker function handles a request of type HASHNEW. It lets the 
   client load up data about a new K-orbit. This request type is used
   for the hash server which is responsible for the K-orbit. That
   means that it is the first one having to store points of this
   K-orbit.  If this K-orbit is not yet known, a new K-orbit-number is
   fetched from the jobserver, which learns in this way of the
   existence of this new K-orbit. If this K-orbit is known (because
   some other client already loaded this data up), the answer is the
   negative of the K-orbit number. The client should not load up the
   rest of the K-orbit to other hash servers in this case. The request
   data looks as follows: First comes the total number of points in
   the K-orbit. After that, the number of representative points
   follows and then the number of further minimal points. After that
   follow first the representative points and then the other minimal
   points. So the total length of the request is 3*sizeof(long) +
   (nrrep+nrmin)*dc_length_pt . */
{
  long nr;              /* number of representative points */
  long nrm;             /* number of further minimal points */
  long i;               /* counter */
  long k;               /* hash key */
  long olen;            /* the total length of the K-orbit */
  void *v;              /* a pointer to the points */
  char *vv;
  long *p;              /* to read longs out of the buffer */
  long t[4];            /* buffer to write numbers to .nrs file */
  long newnumber;       /* a new number for the new K-orbit */

  p = (long *) (question.buf);
  olen = *p++;
  nr = *p++;
  nrm = *p++;

  printf("Getting K-orbit: "); fflush(stdout);
  /* these points must be kept for hash table, we copy them out of
     the buffer: */
  /* Note that in this query nr+nrm is always positive! */
  v = malloc(dc_length_pt * (nr+nrm));
  if (v == NULL) {
    fprintf(stderr,"No more memory for vectors!\n");
    return -1;   /* this is fatal! */
  }
  memcpy(v,p,dc_length_pt * (nr+nrm));
  
  k = ValueHT(hG, v, dc_length_pt);  /* we look up the first one */
  if (k == HTFAIL) {
    /* orbit is new - enter in hash table */
    /* first we fetch a new number from the jobserver: */
    /* note that we use the answer buffer to avoid another malloc! */
    
    if (cs_ensure_bufferlength(answer,sizeof(long))) {
      fprintf(stderr,"No more memory for jobserver request buffer!\n");
      free(v);
      return -1;
    }
    answer->len = 0;
    if (cs_request(jobserver,ORBITNUM,*answer,answer) < 0) {
      /* we are in trouble here! */
      fprintf(stderr,"Bad or no answer from jobserver!\n");
      exit(1);
    }
    newnumber = *((long *)(answer->buf));

    printf("new, #%ld, length %ld, %ld parts, %ld "
	   "additional minimals.\n", newnumber, olen, nr, nrm); fflush(stdout);
 
    /* now we enter those points into the hash: */
    /* note that v is already initialized */
    for (i = nr+nrm, vv=(char*)v; i != 0; i--) {
      AddHT(hG, vv, dc_length_pt, newnumber);
      vv += dc_length_pt;
    }
    /* save vectors in file: */
    if ((signed long) dc_length_pt * (nr+nrm) != 
	write(fdvec, (char *)v, dc_length_pt * (nr+nrm))) {
      fprintf(stderr,"Cannot save new vectors to file, exiting!\n");
      exit(1);
    }

    /* and also store the values: 
       number of representatives,
       pointer to representatives */
    AppendZerosLongList(KOrbRepLen, newnumber+1);
    AppendZerosPtrList(KOrbRep, newnumber+1);
    KOrbRepLen->list[newnumber] = nr;
    KOrbRep->list[newnumber] = v;

    /* save numbers to .nrs file: */
    t[0] = newnumber; t[1] = nr;  t[2] = nrm; t[3] = olen;
    if (4 * sizeof(long) != write(fdnrs, (char *)t, 4 * sizeof(long))) {
      fprintf(stderr,"Cannot save new numbers to file, exiting!\n");
      exit(3);
    }

    /* answer with new orbit number (answer is already long enough!): */
    answer->len = sizeof(long);
    *((long *)(answer->buf)) = newnumber;
  } else {   /* we already know this orbit */
    /* throw them away */
    printf("already known, is number %ld\n", k); fflush(stdout);
    free(v);
    if (cs_ensure_bufferlength(answer,sizeof(long))) {
      /* we are in trouble here! */
      fprintf(stderr,"No memory for answer!\n");
      exit(1);
    }
    answer->len = sizeof(long);
    *((long *)(answer->buf)) = -k;   
    /* the negative of the number of the already existing K-orbit */
  }
  fflush(stdout);
  return 0;
}

long hashnewnr(CS_BUFTYPE question,CS_BUFTYPE *answer) 
/* This worker function handles a request of type HASHNEWNR. It lets the 
   client load up data about a new K-orbit. This request type is used
   after some points are already loaded up to the first hash server
   and a new number for this K-orbit is chosen. Normally the hash
   server will not yet know points from this new K-orbit. Otherwise it
   is a serious error and leads to termination of the hash server. The
   answer is just one long: 0 for OK and -1 for an error.  The request
   data looks as follows: First comes the total number of points in
   the K-orbit. After that, the number of representative points
   follows and then the number of further minimal points. After that
   follow first the representative points and then the other minimal
   points. At last follows a long for the number of this K-orbit. So
   the total length of the request is 4*sizeof(long) +
   (nrrep+nrmin)*dc_length_pt . */
{
  long nr;              /* number of representative points */
  long nrm;             /* number of further minimal points */
  long i;               /* counter */
  long k;               /* hash key */
  long olen;            /* the total length of the K-orbit */
  void *v;              /* a pointer to the points */
  char *vv;
  long *p;              /* to read longs out of the buffer */
  long t[4];            /* buffer to write numbers to .nrs file */
  long newnumber;       /* a new number for the new K-orbit */

  p = (long *) (question.buf);
  olen = *p++;
  nr = *p++;
  nrm = *p++;
  /* here at the end is the K-orbit number: */
  memcpy(&newnumber,(char *) p + dc_length_pt * (nr+nrm),sizeof(long));

  printf("Getting K-orbit (%ld): ", newnumber); fflush(stdout);
  /* these vectors must be kept for hash table, we copy them out of
     the buffer: */
  if (nr+nrm > 0)
    v = malloc(dc_length_pt * (nr+nrm));
  else
    v = malloc(sizeof(long));

  if (v == NULL) {
    fprintf(stderr,"No more memory for vectors!\n");
    return -1;   /* this is fatal! */
  }
  /* to ensure that lists are long enough when entry is checked */
  AppendZerosLongList(KOrbRepLen, newnumber+1);
  AppendZerosPtrList(KOrbRep, newnumber+1);
  if (nr+nrm > 0) {
    memcpy(v,p,dc_length_pt * (nr+nrm));
    k = ValueHT(hG, v, dc_length_pt);  /* we look up the first one */
  } else {
    /* In this case the hash will reveal nothin */
    if (KOrbRep->list[newnumber] == NULL)  /* a new K-orbit as expected */
      k = HTFAIL;
    else
      k = HTFAIL-1;   /* This only has to be unequal to HTFAIL */
  }
  if (k == HTFAIL) {
    /* now we enter those points into the hash: */
    /* note that v is already initialized */
    for (i = nr+nrm, vv=(char*)v; i != 0; i--) {
      AddHT(hG, vv, dc_length_pt, newnumber);
      vv += dc_length_pt;
    }

    printf("new, #%ld, length %ld, %ld parts, %ld "
	   "additional minimals.\n", newnumber, olen, nr, nrm); fflush(stdout);
    
    /* save vectors in file: */
    if (nr+nrm > 0 &&       /* do not save 0 bytes! */
	(signed long) dc_length_pt * (nr+nrm) != 
	write(fdvec, (char *)v, dc_length_pt * (nr+nrm))) {
      fprintf(stderr,"Cannot save new vectors to file, exiting!\n");
      exit(1);
    }

    /* and also store the values: 
       number of representatives,
       pointer to representatives */
    KOrbRepLen->list[newnumber] = nr;
    KOrbRep->list[newnumber] = v;

    /* save numbers to .nrs file: */
    t[0] = newnumber; t[1] = nr;  t[2] = nrm; t[3] = olen;
    if (4 * sizeof(long) != write(fdnrs, (char *)t, 4 * sizeof(long))) {
      fprintf(stderr,"Cannot save new numbers to file, exiting!\n");
      exit(3);
    }

    /* answer with 0: */
    if (cs_ensure_bufferlength(answer,sizeof(long))) {
      fprintf(stderr,"No more memory for jobserver request buffer!\n");
      free(v);
      return -1;
    }
    answer->len = sizeof(long);
    *((long *)(answer->buf)) = newnumber;
    fflush(stdout);
    return 0;
  } else {   /* we already know this orbit, this is impossible! */
    /* throw them away */
    fprintf(stderr,"already known, is number %ld\n", k);
    exit(1);
  }
  return 0;  /* this is only to make some stupid compiler happy that does
		not know the meaning of `exit' */
}

long hashorb(CS_BUFTYPE question,CS_BUFTYPE *answer)
/* This request type returns the representatives of a K-orbit which should
   already be known to the hash server. The question contains just one
   long, which is the positive K-orbit number. The answer consists of one long 
   for the number of points and then the representatives. It is possible
   that the hash server is asked for a K-orbit it does not know yet (if one 
   client is loading up this K-orbit and has not yet reached this server
   and another client already asks everybody for this K-orbit). In this
   case only one long containing a negative number is returned. */
{
  long num;   /* number of K-orbit */
  long nr;    /* number of representatives in K-orbit */

  num = *((long *)(question.buf));
  if (KOrbRep->len <= num || KOrbRep->list[num] == NULL) {
    /* we do not know this orbit, this is strange, but possible! */
    if (cs_ensure_bufferlength(answer,sizeof(long)) < 0) {
      /* Houston, we have a problem! */
      fprintf(stderr,"Cannot allocate memory for answer!\n");
      exit(1);
    }
    fprintf(stderr,"Note: We got a HASHORB request for an unknown K-orbit!\n");
    *((long *)(answer->buf)) = -1;
    answer->len = sizeof(long);
    return 0;
  }
							     
  nr = KOrbRepLen->list[num];
  if (cs_ensure_bufferlength(answer,sizeof(long) + dc_length_pt * nr) < 0) {
    /* No more memory! */
    fprintf(stderr,"Cannot allocate memory for answer!\n");
    exit(1);
  }
    
  *((long *)(answer->buf)) = nr;  /* note that this can be zero! */
  answer->len = sizeof(long) + dc_length_pt * nr;
  if (nr > 0)
    memcpy((long *)(answer->buf) + 1,KOrbRep->list[num],dc_length_pt * nr);

  printf("Sending %ld representatives for K-orbit %ld.\n", nr, num); 
  fflush(stdout);
  return 0;
}


/* Here is the table of our services: */
CS_WORKER hashworkerlist[] = { { HASHQUERY, hashquery },
			       { HASHNEW,   hashnew },
			       { HASHNEWNR, hashnewnr },
			       { HASHORB,   hashorb } };


int main(int argc,const char *argv[])
/* This program uses the "clntserv.c" library, being a server. So
   after initialization there is a main loop and all the functionality
   is only in callback functions. */
{
  unsigned short int port;    /* port number for incoming connections */
  unsigned short int jobservport;    /* port of job server */
  /* file names for saving hash table: */
  char filevec[_POSIX_PATH_MAX+5];    /* file name for vectors file */
  char filenrs[_POSIX_PATH_MAX+5];    /* file name for numbers file */
  /* Note that +4 is for the appendices .vec and .nrs resp. */
  FILE *input;         /* to read the input file */
  char buffer[1024];   /* to read the input file */
  long len;            /* length of input number list */
  long i;              /* counter */

  /* read commandline:   
     dchashserv inputfile port jobserverhost jobserverport hashdata */
  if (argc != 6) {
    puts("Usage: dchashserv inputfile port jobserverhost jobserverport "
	 "hashdata");
    puts("Where: inputfile     : is the filename of this file");
    puts("       port          : is the port to offer the service");
    puts("       jobserverhost : hostname of the jobserver");
    puts("       jobserverport : port of the jobserver");
    puts("       hashdata      : file name for the two hash data files,");
    puts("                       .vec and .nrs is appended resp.");
    exit(0);
  }

  port = atoi(argv[2]);
  jobservport = atoi(argv[4]);
  strncpy(filevec,argv[5],_POSIX_PATH_MAX); filevec[_POSIX_PATH_MAX] = 0;
  strcat(filevec,".vec");
  strncpy(filenrs,argv[5],_POSIX_PATH_MAX); filenrs[_POSIX_PATH_MAX] = 0;
  strcat(filenrs,".nrs");

  /* First we read the input file: */
  input = fopen(argv[1],"r");
  if (input == NULL) {
    fprintf(stderr,"Cannot open input file %s",argv[1]);
    perror(":");
    exit(1);
  }

  /* read the configuration numbers: */
  dcreadline(buffer,1023,input);   /* this reads the number of numbers */
  len = atol(buffer);
  hashservconf = NewLongList(len);   /* Allocate configuration number list */
  if (hashservconf == NULL) {  /* Problems with allocation */
    perror("Cannot allocate configuration number list:");
    exit(1);
  }
  hashservconf->len = len;
  for (i = 0;i < len;i++) {
    dcreadline(buffer,1023,input);   /* read another number */
    hashservconf->list[i] = atol(buffer);
  }

  fclose(input);

  /* We store those for convenience in separate variables: */
  hGlen = hashservconf->list[0];

  /* Initialize our global data: */
  KOrbRepLen = NewLongList(1000L);
  KOrbRep = NewPtrList(1000L);
  hG = NewHT(hGlen, DC_HASHFUN, DC_CMP_PTS);

  DC_HASHSERVINIT(hashservconf);

  /* Now we need a connection to the job server: */
  jobserver = cs_connect((char *)(argv[3]),jobservport);
  if (jobserver < 0) {
    fprintf(stderr,"Cannot contact job server!\n");
    exit(1);
  }

  /* open files and read already stored information: */
  if (( fdnrs = open(filenrs, O_RDWR | O_CREAT | O_APPEND | O_SYNC, 
		     0644)) == -1) {
      fprintf( stderr, "Cannot open file %s ", filenrs);
      perror(":");
      exit(1);
  }
  if (( fdvec = open(filevec, O_RDWR | O_CREAT | O_APPEND | O_SYNC,
		     0644)) == -1) {
      fprintf( stderr, "Cannot open file %s ", filevec);
      perror(":");
      exit(1);
  }
  ReadKnownOrbits();

  cs_start_server(hashworkerlist,port);
  return 0;
}
