/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * dcgeneric/dcjobserv.c   The job server, which controls everything.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This program organizes the whole calculation by "serving" jobs to be
 * calculated.
 * 
 * $Id: dcjobserv.c,v 1.2 2000/04/28 09:26:46 neunhoef Exp $
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "clntserv.h"
#include "dcutil.h"
#include "dcservices.h"


/* most variables will be global in the server such that they can be
   accessed from all service handlers */
static FILE *cmat;                 /* file handle for result */
static LongList *jobrow;           /* for job queue */
static long pos = 0;               /* helper in newjob for rescheduling */

/************   handlers for the different services   ********************/

/* NEWJOB - returns job numbers, if there are at least 20 jobs, all jobs 
   are given away and some are not yet returned, then we schedule such jobs 
   again.
   A return value -1 says that all jobs are done.
   A return value -2 tells the client to wait some seconds before asking
   again (during the startup phase).    */
long newjob(CS_BUFTYPE question, CS_BUFTYPE *answer)
{
  long i;
  question = question;  /* unused - just to avoid compiler warning */

  cs_ensure_bufferlength(answer, sizeof(long));
  answer->len = sizeof(long);

  /* entries in `jobrow': 0 not handled, 1 given to client, 2 got
     result */
  for (i = 0; i < jobrow->len && jobrow->list[i] > 0; i++);
  if (i == jobrow->len && i>19) { /* all known jobs handled and at least 20
				     jobs in list */
    for (i = pos+1; i != pos && jobrow->list[i] != 1; i = (i+1) % jobrow->len);
    if (jobrow->list[i] == 1) { /* schedule this job again */
      pos = i;
      printf("Scheduling job %ld again.\n", i); fflush(stdout);
      *((long *)answer->buf) = i;
      return 0;
    } 
    else { /* all done */
      printf("No more jobs, ready!\n"); fflush(stdout);
      *((long *)answer->buf) = -1;
      return -1;
    }
  } 
  else if (i == jobrow->len) { /* not enough jobs in queue, client
				  should wait */
    printf("No job available, please wait.\n"); fflush(stdout);
    *((long *)answer->buf) = -2;
    return 0;
  }
  else {  /* job found which is not yet handled */
    printf("Scheduling job %ld.\n", i); fflush(stdout);
    if (i==0) jobrow->list[0] = 2;
    else jobrow->list[i] = 1;
    *((long *)answer->buf) = i;
    return 0;
  } 
}

/* ORBITNUM - just return the next number and add entry to `jobrow' */
long orbitnum(CS_BUFTYPE question, CS_BUFTYPE *answer)
{
  question = question;  /* unused - just to avoid compiler warning */

  cs_ensure_bufferlength(answer, sizeof(long));
  answer->len = sizeof(long);

  /* create new job with status 0 */
  AddLongList(jobrow, 0L);
  *((long *)answer->buf) = jobrow->len - 1;
  printf("Found orbit number %ld.\n", jobrow->len - 1); fflush(stdout);
  return 0;
}

/* GETRESULT - get result of a job, the rows of the condensation matrices 
   data in question consists of list of longs:
     nr of job; 
     nr of rows; 
     length of row 1; 
     ..entries of row 1..; 
     length of row 2; 
     .....
   answer: 0L
*/
long getresult(CS_BUFTYPE question, CS_BUFTYPE *answer)
{
  long jobnr, ng, rowlen, *ptr, k, i;
  
  cs_ensure_bufferlength(answer, sizeof(long));
  answer->len = sizeof(long);
  
  ptr = (long *) question.buf;
  jobnr = *(ptr++); /* job number = row index of result */
  if (jobrow->list[jobnr] < 2) {  /* not yet known */
    ng = *(ptr++);    /* number of elements to condense */
    
    for (k=0; k<ng; k++) {
      /* writing rows in GAP file, e.g.: "cmat.1[13]:=[1,2,3];" */
      rowlen = *(ptr++);
      fprintf(cmat, "cmat.%ld[%ld]:=[%ld", k + 1, jobnr, *(ptr++));
      for (i = 1; i < rowlen; i++) {
	fprintf(cmat, ",%ld", *(ptr++));
      }
      fprintf(cmat, "];\n");
    }
    fflush(cmat);
    /* mark job as finished */
    jobrow->list[jobnr] = 2;
  }
  *((long *) answer->buf) = 0;
  printf("Got result rows from job %ld.\n", jobnr); fflush(stdout);
  return 0;
}

/* the main program is now very simple */
int main(int argc, char *argv[])
{
  unsigned short int port;
  long nrknownorbits, i;
  CS_WORKER handlers[] = { { NEWJOB, newjob },
			   { GETRESULT, getresult },
			   { ORBITNUM, orbitnum },
			   { 0,   NULL } };
  
  /* evaluate command line */
  if (argc < 2 || argc > 4) {
    puts("Usage: dcjobserv port resultfile [nrknownorbits]");
    exit(1);
  }

  /* port number for connections */
  port = atoi(argv[1]);

  /* open file for results */
  if (( cmat = fopen(argv[2], "w")) == NULL) {
    perror( "Cannot open result file:");
    exit(1);
  } 
  else {
    /* we store result in GAP readable record, which may be 
       initialized first */
    fprintf(cmat, "if not IsBound(cmat) then cmat := rec(); fi;\n");
  }


  /* a hint in case of new startup */
  if (argc == 4)
    nrknownorbits = atol(argv[3]);
  else
    nrknownorbits = 0;
  /* ------------------------- command line read  -----------------*/

  /* allocate list for managing jobs, see `newjob' */
  jobrow = NewLongList(1000L);
  for (i=0; i<=nrknownorbits; i++)
    AddLongList(jobrow, 0L); 

  /* here the work loop begins */
  cs_start_server(handlers, port);
  return 0;
}
