/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * dcgeneric/dcutil.c   Utilities for the generic programs.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This file contains some helper routines, especially for hash lookups
 * and list processing.
 * 
 * $Id: dcutil.c,v 1.5 2000/04/28 09:26:30 neunhoef Exp $
 *
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include "dcutil.h"

#ifdef DOHASHSTATISTICS
long long nraccesses = 0;
long long nrhits = 0;
long long nrsteps = 0;
#endif

/* basic functions for hash tables */

/* initializes new hash table with and w/o space for values 
   input: length l, hash function f (f must give long values) and
   compare function for two elements (given by void pointers and
   their length */
HashType* NewHT( long len, long (*hfun)(void *, size_t), 
		 int (*cmp)(void *, void *, size_t))
{
  HashType * res;
  res = (HashType *)malloc(sizeof(HashType));
  res->len = len;
  res->nr = 0;
  res->els = (void **)calloc(len, sizeof(void *));
  res->vals = (long *)malloc(len * sizeof(long));
  res->hfun = hfun;
  res->cmp = cmp;
  return res;
}

HashType *NewHTNoVals( long len, long (*hfun)(void *, size_t),
		       int (*cmp)(void *, void *, size_t))
{
  HashType *res;
  res = (HashType *)malloc(sizeof(HashType));
  res->len = len;
  res->nr = 0;
  res->els = (void **)calloc(len, sizeof(void *));
  res->vals = NULL;
  res->hfun = hfun;
  res->cmp = cmp;
  return res;
}

/* get a value (always 1 for NoVals table) from hash table or 
`HTFAIL' if not known */
long ValueHT(HashType *ht, void *x, size_t len)
{
  long h;
  /* get hash value */
  h = (unsigned long)(ht->hfun(x, len)) % ht->len;
  for (;ht->els[h] != NULL && ht->cmp(x, ht->els[h], len) != 0; ){
    h++;
    if (h == ht->len) h = 0;
  }
  if (ht->els[h] == NULL){
    return HTFAIL;
  }
  else {
    if (ht->vals == NULL) {
      return 1L;
    }
    else {
      return ht->vals[h];
    }
  }
}

/* add an entry to a hash table */
/* exits the program if no slot available */
long AddHT(HashType *ht, void *x, size_t len, long val)
{
  long h;

#ifdef DOHASHSTATISTICS
  nraccesses++;
#endif

  if (ht->len == ht->nr + 1) /* return HTFAIL; */ {
    printf("Error: Hash table too small, only %ld entries allowed.\n", 
	   ht->len - 1);
    exit(2);
  }
  h = (unsigned long)(ht->hfun(x, len)) % ht->len;
  for (;ht->els[h] != NULL; ){

#ifdef DOHASHSTATISTICS
  nrsteps++;
#endif

    h++;
    if (h == ht->len) h = 0;
  }
  ht->els[h] = x;
  if (ht->vals != NULL)
    ht->vals[h] = val;
  ht->nr++;
  
#ifdef DOHASHSTATISTICS
  if (nraccesses % 100000 == 0)
    printf("Hash statistics: \n%lld stores, %lld steps"
           "\nNumber of vectors in Hash: %ld(of %ld).\n",
           nraccesses, nrsteps, ht->nr, ht->len);
#endif

  return h;
}

/* in case the hash value `hval' is already known */
long AddHTHVal(HashType *ht, void *x, long hval, long val)
{
  long h;

#ifdef DOHASHSTATISTICS
  nraccesses++;
#endif

  if (ht->len == ht->nr + 1) /* return HTFAIL; */ {
    printf("Error: Hash table too small, only %ld entries allowed.\n", 
	   ht->len - 1);
    exit(2);
  }
  h = hval;
  for (;ht->els[h] != NULL; ){

#ifdef DOHASHSTATISTICS
  nrsteps++;
#endif

    h++;
    if (h == ht->len) h = 0;
  }
  ht->els[h] = x;
  if (ht->vals != NULL)
    ht->vals[h] = val;
  ht->nr++;
  
#ifdef DOHASHSTATISTICS
  if (nraccesses % 100000 == 0)
    printf("Hash statistics: \n%lld stores, %lld steps"
           "\nNumber of vectors in Hash: %ld(of %ld).\n",
           nraccesses, nrsteps, ht->nr, ht->len);
#endif

  return h;
}

/* clearing a hash table for reuse */
void ClearHT(HashType *ht) 
{
  long h;
  void **p;
  p = ht->els;
  for (h = 0; h < ht->len; h++) {
    p[h] = NULL;
  }
  ht->nr = 0;
}

LongList *NewLongList(long slen) 
{
  LongList *l;
  l = (LongList *)malloc(sizeof(LongList));
  l->maxlen = slen;
  if (l->maxlen < 1) l->maxlen = 1;
  l->len = 0;
  l->list = (long *)malloc(l->maxlen*sizeof(long));
  return l;
}

long AddLongList(LongList *l, long v)
{
  if (l->maxlen == l->len) {
    /* make list by factor approx. a factor 3/2 bigger, if no more space */
    l->maxlen = l->maxlen*3/2+1;
    l->list = (long *)realloc(l->list, l->maxlen*sizeof(long));
  }
  l->list[l->len] = v;
  (l->len)++;
  return l->len;
}

long AppendZerosLongList(LongList *l, long nlen)
/* Ensures that the list l has enough room for at least `nlen'
   longs and length of at least `nlen' longs. This is accomplished by
   `realloc'ing the buffer and appending zeros if necessary. If
   something goes wrong, a negative number is returned and 0
   otherwise. */
{
  long newlen;
  long *v;
  long i;

  if (l->maxlen < nlen) {  /* we have to do something! */
    newlen = l->maxlen*3/2+1;
    if (newlen < nlen) newlen = nlen;
    l->list = (long *) realloc(l->list,newlen * sizeof(long));
    if (l->list == NULL) return -1;
    l->maxlen = newlen;
  }
  if (l->len < nlen) {
    for (i = l->len, v = l->list + l->len;i < nlen;i++)
      *v++ = 0;
    l->len = nlen;
  }
  return 0;
}

/* Things to work with lists of pointers: */

PtrList *NewPtrList(long slen) 
/* Allocates a new PtrList for at most `slen' pointers. The length is
   set to 0. Returns pointer to a PtrList struct. Terminates, if
   something goes wrong. */
{
  PtrList *l;
  l = (PtrList *) malloc(sizeof(PtrList));
  if (l == NULL) {
    perror("Could not allocate PtrList:");
    exit(1);
  }
  l->maxlen = slen;
  if (l->maxlen < 1) l->maxlen = 1;
  l->len = 0;
  l->list = (void **) malloc(l->maxlen * sizeof(void *));
  if (l->list == NULL) {
    perror("Could not allocate PtrList:");
    exit(1);
  }
  return l;
}

long AddPtrList(PtrList *l, void *v)
/* Adds a pointer at the end of the PtrList *l. If this list has not
   enough room, it is enlarged by a factor of 3/2 by realloc. If this
   does not work, the program is terminated. The new length is
   returned. */
{
  if (l->maxlen == l->len) {
    /* make list by factor approx. a factor 3/2 bigger, if no more space */
    l->maxlen = l->maxlen*3/2+1;
    l->list = (void **) realloc(l->list, l->maxlen * sizeof(void *));
    if (l->list == NULL) {
      fprintf(stderr,"Could not reallocate PtrList!\n");
      exit(1);
    }
  }
  l->list[l->len] = v;
  (l->len)++;
  return l->len;
}

long AppendZerosPtrList(PtrList *l, long nlen)
/* Ensures that the list l has enough room for at least `nlen'
   pointers and length of at least `len' pointers. This is
   accomplished by `realloc'ing the buffer and appending zero pointers
   if necessary. If something goes wrong, a negative number is
   returned and 0 otherwise. */
{
  long newlen;
  void **v;
  long i;

  if (l->maxlen < nlen) {  /* we have to do something! */
    newlen = l->maxlen*3/2+1;
    if (newlen < nlen) newlen = nlen;
    l->list = (void **) realloc(l->list,newlen * sizeof(void *));
    if (l->list == NULL) return -1;
    l->maxlen = newlen;
  }
  if (l->len < nlen) {
    for (i = l->len, v = l->list + l->len;i < nlen;i++)
      *v++ = NULL;
    l->len = nlen;
  }
  return 0;
}

void dcreadline(char *b,size_t l,FILE *f)
/* Reads a line from file f into buffer b which has space for l
   bytes. Ignores everything after the first # character and ignores
   empty lines consisting only of whitespace. */
{
  int done = 0;
  char *p;

  while (!done) {
    if (fgets(b,l,f) == NULL) {   /* this is an error */
      perror("Could not read input file:");
      exit(1);
    }
    p = strchr(b,'#');
    if (p) *p = 0;
    for (p = b;*p == ' ' || *p == '\t';p++);
    if (*p) /* we found non-whitespace */
      done = 1;
  }
}

ssize_t dcreadfile(FILE *file,char **buffer)
/* Reads a text file into memory. The file is read completely in one block.
   Then in every line everything from the first # sign on is deleted. The
   memory needed is fetched via `malloc'. A pointer on the remaining text
   is stored in *buffer. After removal of comments the memory is realloced
   to the actuall length. The number of bytes in the buffer (can be 0) is 
   returned or -1 in case of an error. The idea is to use `dcfindtoken'
   and `sscanf' to grab the values. */
{
  struct stat statbuf;  /* to determine the length of the file */
  char *buf;            /* points to buffer */
  long res;             /* result */
  char *p,*q;           /* to walk along the string */
  char *bufferend;      /* end of buffer */

  res = fstat(fileno(file), &statbuf);
  if (res < 0) return -1;

  buf = malloc(statbuf.st_size+1);
  if (buf == NULL) return -1;

  res = fread(buf,1,statbuf.st_size,file);
  if (res < statbuf.st_size) {
    free(buf);
    return -1;
  }

  bufferend = buf + statbuf.st_size;

  /* Now we remove comments: */
  p = buf;  /* reading pointer */
  q = buf;  /* writing pointer */
  while (p < bufferend) {
    if (*p == '#') {  /* delete everything up to the next newline */
      while (p < bufferend && *p != '\n') p++;
    }
    *q++ = *p++;
  }

  /* A terminating zero: */
  *q = 0;
  *buffer = realloc(buf,(q - buf) + 1);
  if (*buffer == NULL)
    return -1;

  return (q - buf);
}

char *dcfindtoken(char **buf,char *bufend)
/* First advances *buf over all whitespace (spaces, tabs, newlines), then
   stores this position, then advances *buf further up to the next whitespace
   character, places a zero character, and advances *buf one more character.
   returns stored position. *buf is never advanced to a value greater than
   bufend. Assumes that it is possible (and safe) to place a zero byte at
   the address bufend points to. Each call to `dcfindtoken' returns a pointer 
   to a zero-terminated string containing exactly one token. */
{
  /* Skip whitespace: */
  char *store;

  while (*buf < bufend && isspace(**buf))
    (*buf)++;
  store = *buf;
  while (*buf < bufend && !isspace(**buf))
    (*buf)++;
  *(*buf) = 0;
  if (*buf < bufend)
    (*buf)++;

  return store;
}
