/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * include/clntserv.h   header for the "clntserv" library for client/server com
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This is the header file for the "clntserv" library for client/server
 * communications.
 * 
 * $Id: clntserv.h,v 1.2 2000/04/28 09:44:19 neunhoef Exp $
 */

#include <stddef.h>
#include <unistd.h>


/* Constants: */

#define CS_MAXSERVERS 200
#define CS_MAXCLIENTS 200
#define CS_BLOCKING_TIMEOUT 300
#define CS_MAXQUEUE 10

/* Configuration of behaviour: */


/* If this variable is non-zero, most errors lead to an exit(1) call: */
/* The default is 1. */
extern int cs_exit_on_error;

/* If this variable is non-zero, the library automatically tries to do some 
   input/output via `cs_do_io' whenever a non-blocking call is issued.
   The default is 1. */
extern int cs_automatic_io;

/* The following value in seconds is the timeout for blocking I/O. Set to
   zero if you do not want any timeout. The default value is the 
   abovementioned constant CS_BLOCKING_TIMEOUT. */
extern int cs_blocking_timeout;

/* New types: */

typedef struct cs_buftype {
  void *buf;       /* start address */
  size_t buflen;   /* length of the buffer */
  size_t len;      /* length of the data in the buffer */
} CS_BUFTYPE;


typedef struct cs_reqtype {
  struct cs_reqtype *prev; /* doubly linked list, first has NULL here */
  struct cs_reqtype *next; /* last has NULL here */
  CS_BUFTYPE question;     /* The input, this is never changed */
  size_t alreadysent;      /* Amount of data already sent */
  CS_BUFTYPE answer;       /* The output, this may be changed, see
			      functions */
  size_t alreadyrecv;      /* Amount of data already received */
  int allocateanswer;      /* 0 : do not touch answer buffer, 1 : we will
			      allocate an answer buffer big enough, 2 : we
			      allocated the answer buffer and have to free
			      it when forgetting this request */
  int status;              /* 0  : header not fully sent,  1 : header sent,
			      but question  not yet, 2:  question sent, but
			      answer header  not fully  received  yet,  3 :
			      answer header   received, data not  yet,  4 :
			      completed, -1 : completed with error */
  long server;             /* Number of server */
  long header[2];          /* Buffer for reading/writing of header */
  long answerlength;       /* length of answer (sent as header) */
  int cancelled;           /* Flag, whether the request was cancelled */
  int stopped;             /* Flag, whether the request is stopped */
  void *private;           /* Pointer for usage by the user */
} CS_REQTYPE;

typedef long (* CS_WORKERTYPE) (CS_BUFTYPE question,CS_BUFTYPE *answer);
/* This is the type for a worker function. It gets its input in the
   buffer `question' and has to allocate a new answer buffer via
   `realloc'. This will *not* be freed automatically later on, when
   all the data is sent to the client. The next call to the worker
   will again be able to use this buffer. This function does not have
   to worry about input and output. If the return value is negative,
   the client will be thrown out. A return value of 0 indicates
   success. */

typedef struct cs_worker { long magic; CS_WORKERTYPE worker; } CS_WORKER;

/* Global data in the client: */

/* Beginning of doubly linked list of req.: */
extern CS_REQTYPE *cs_req_first;
/* End of doubly linked list of requests: */
extern CS_REQTYPE *cs_req_last;    
/* Total number of bytes in question buffers of requests: */
extern size_t cs_question_bytes;
/* Total number of bytes in answer buffers of requests: */
extern size_t cs_answer_bytes;
/* Total number of requests in queue: */
extern long cs_number_requests;
/* Total number of answered (successful or with error!) requests in queue: */
extern long cs_number_answered;


/* Some useful things: */

ssize_t readbyselect(int fd, void *buf, size_t count, int skipselect);
/* This function implements non-blocking input by using `select': It tries
   to read up to `count' bytes from the file descriptor `fd' into the
   buffer pointed to by `buf'. It returns the number of bytes actually
   read (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). It first calls `select'
   with a timeout of zero and tries to read count bytes. This is
   repeated until either there is no more data present, the connection
   is broken or `count' bytes have been read. If the flag `skipselect'
   is !=0, we skip the first select assuming that `select' was called
   right before the call to this function. */

ssize_t writebyselect(int fd, void *buf, size_t count, int skipselect);
/* This function implements non-blocking output by using `select': It tries
   to write up to `count' bytes to the file descriptor `fd' from the
   buffer pointed to by `buf'. It returns the number of bytes actually
   written (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). It first calls `select'
   with a timeout of zero and tries to write count bytes. This is
   repeated until either the file descriptor cannot take more data,
   the connection is broken or `count' bytes have been written. If the
   flag `skipselect' is !=0, we skip the first select assuming that
   `select' was called right before the call to this function. */

ssize_t readblock(int fd, void *buf, size_t count);
/* This function implements blocking input by using `select': It tries
   to read up to `count' bytes from the file descriptor `fd' into the
   buffer pointed to by `buf'. It returns the number of bytes actually
   read (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). If necessary, multiple `read's are
   done but if a timeout occurs, the function returns. */

ssize_t writeblock(int fd, void *buf, size_t count);
/* This function implements blocking output by using `select': It tries
   to write up to `count' bytes to the file descriptor `fd' from the
   buffer pointed to by `buf'. It returns the number of bytes actually
   written (which can be 0) or -1 if there is an error (for example a
   closed connection if `fd' is a socket). If necessary, multiple `write's are
   done but if a timeout occurs, the function returns. */

CS_BUFTYPE cs_new_buffer(size_t startlen);
/* This function allocates a new buffer with length startlen and sets
   up the struct for this. It returns a struct with `buf' entry NULL,
   if the allocation does not work. If `cs_exit_on_error' is non-zero,
   the program is terminated instead. If `startlen' is zero, the struct
   returned has `buf' entry NULL which is not an error. This struct can
   be used to call non-blocking requests where the answer buffer is
   allocated automatically. */

int cs_ensure_bufferlength(CS_BUFTYPE *buf,size_t len);
/* This function checks, whether the buffer `buf' is at least `len'
   bytes long, if not, `realloc' is called. Returns 0 if this worked
   and -1 otherwise. */


/* For the client: */

long cs_connect(char *name,unsigned short int port);
/* Opens connection to a server. `name' ist a internet hostname,
   `port' a TCP port number. The result is an identifier for this
   connection (non-negative number) if everything is OK and otherwise
   a negative number. If the global flag `cs_exit_on_error' is
   non-zero, this function does not return if the connection is not
   established, but calls `exit(1)'. */

long cs_request(long server, long magic, 
		CS_BUFTYPE question, CS_BUFTYPE *answer);
/* This function sends a request and blocks until the answer has
   arrived. `server' is a server identifier as returned by
   `cs_connect', `magic' is a number which specifies the service
   required from the server. `question' specifies the input for the
   request. `*answer' specifies a buffer for the reply, which must not
   be longer than `answer->buflen'. `answer->len' is modified
   according to the length of the result. No memory allocation happens
   within this function. If everything is OK, 0 is returned and
   otherwise a negative number. If the global flag `cs_exit_on_error'
   is non-zero, this function does not return if the connection is not
   established, but calls `exit(1)'. */

long cs_request_malloc(long server, long magic, CS_BUFTYPE question,
		       CS_BUFTYPE *answer);
/* This function sends a request and blocks until the answer has
   arrived. `server' is a server identifier as returned by
   `cs_connect', `magic' is a number which specifies the service
   required from the server. `question' specifies the input for the
   request. `*answer' specifies a buffer for the reply. If
   `answer->buf' is NULL, a new buffer of the correct length is
   malloced. Otherwise, `*answer' specifies an already existing
   buffer, which must have been allocated by `malloc'. The function is
   allowed to issue `realloc', if the buffer is not long enough for
   the answer. In this case `answer->buflen' is modified
   accordingly. `realloc' is not called if the buffer is too long for
   the answer. However `answer->len' is modified according to the
   length of the result. The result is 0 if everything was OK and a
   negative number if anything was wrong.  If the global flag
   `cs_exit_on_error' is non-zero, this function does not return if
   the connection is not established, but calls `exit(1)'. Note that
   it is the responsibility of the caller to free the answer buffer
   once it is no longer needed. */

CS_REQTYPE *cs_request_nb(long server, long magic, CS_BUFTYPE question,
		          CS_BUFTYPE answer, void *private);
/* This function handles a request. It is however a non-blocking
   version. Use `cs_isready' and `cs_whichready' to ask for results.
   `server' is a server identifier as returned by `cs_connect',
   `magic' is a number which specifies the service required from the
   server. `question' specifies the input for the request. `answer'
   specifies a buffer for the reply, which must not be longer than
   `answer.buflen'. `answer.len' is modified according to the length
   of the result. Note that this will only be modified, when the answer
   header is there, which occurs in some call to `cs_isready', `cs_do_io' or
   `cs_whichready'. The modification takes place in the instance of
   CS_BUFTYPE which is stored in the request queue. Do not expect that 
   your original struct is changed! No memory allocation happens within this
   function. The result is a pointer to the queueing data for this
   request. If the global switch `cs_automatic_io' is on, input/output is
   tried after queueing up the request by calling `cs_do_io'
   automatically. `private' is a void pointer that is stored with the
   request and used for nothing. */

CS_REQTYPE *cs_request_nb_malloc(long server, long magic, 
	       CS_BUFTYPE question,CS_BUFTYPE answer,void *private);
/* This function handles a request. It is however a non-blocking version.
   Use `cs_isready' and `cs_whichready' to ask for results.  `server'
   is a server identifier as returned by `cs_connect', `magic' is a
   number which specifies the service required from the server.
   `question' specifies the input for the request.  `answer' specifies
   a buffer for the reply.  If `answer.buf' is NULL, a new buffer of
   the correct length is malloced.  In this case, the corresponding
   `cs_forget' call will free this buffer again.  Otherwise, `answer'
   specifies an already existing buffer, which must have been
   allocated by `malloc'. The function is allowed to issue `realloc',
   if the buffer is not long enough for the answer.  In this case
   `answer.buflen' is modified accordingly.  `realloc' is not called
   if the buffer is too long for the answer.  However `answer.len' is
   modified according to the length of the result.  Note that this
   will only be modified, when the answer header is there, which
   occurs in some call to `cs_isready', `cs_do_io' or
   `cs_whichready'. The modification takes place in the instance of
   CS_BUFTYPE which is stored in the request queue. Do not expect that
   your original struct is changed! `cs_forget' will not free the
   buffer in the case where the original allocation was not done in
   this library.  It is the responsibility of the caller to free the
   answer buffer once it is no longer needed.  If the global switch
   `cs_automatic_io' is on, input/output is tried after queueing up
   the request by calling cs_do_io automatically. `private' is a void
   pointer that is stored with the request and used for nothing. */

void cs_do_io(void);
/* This function does input and output on all open connections handled 
   by this library in a non-blocking way (using select). This can lead
   to the completion of requests, which means, that the answer is
   fully there. It is the responsibility of the user of this library
   to call this often enough, whatever that means... */

int cs_isready(CS_REQTYPE *request);
/* This function checks, whether the request specified by `request' is
   completed, which means, that the answer is fully there. Note that input
   or output is done before that, only if `cs_automatic_io' is non-zero, so
   otherwise you should probably call `cs_do_io' beforehand. It is
   guaranteed that `cs_isready' needs nearly no CPU time. The return value
   is 0, if the request is not yet completed, 1, if it is sucessfully
   completed and negative, if an error occurred. The answer can be found in
   the buffer, which is specified by the exact instance of `CS_BUFTYPE' the
   request was started with. */

CS_REQTYPE *cs_whichready(long server);
/* This function checks, whether there is a request which was sent to
   the server `server', which is already completed. If so, the one
   which was sent first (out of those completed) ist returned. Note
   that no "overtaking" can occur for the requests sent to one
   specific server. If `server' is equal to -1, all requests to all
   servers are considered. Note that in this case "overtaking" can
   occur, it may very well be that one server answers more
   quickly. The function returns the request which was sent first (out
   of those already completed). Note that input or output is done
   before that, only if `cs_automatic_io' is non-zero, so otherwise you
   should probably call `cs_do_io' beforehand. It 
   is guaranteed that `cs_whichready' needs nearly no CPU time. The
   return value is NULL, if the request is not yet completed. */

void cs_forget(CS_REQTYPE *request, int freebuffers);
/* This function is used to clear a completed request out of the data
   structures. The request `request' must be completed and is deleted
   in the doubly linked list. If `freebuffers' is equal to 0, the
   answer buffer is only freed, if this library has allocated it.
   If `freebuffers' is non-zero, the answer buffer and the question
   buffer are freed automatically. */

void cs_stop_nb(CS_REQTYPE *req);
/* Stops the request `req'. This means that its `stopped' flag is set. This 
   can be undone with `cs_continue_nb'. A stopped request for which I/O
   already is started will be completed regardless such that the protocol
   with the server is not broken. However, when the server looks for the
   next request for a server, stopped requests are not considered. For an
   already completed request it means nothing to be stopped. */

void cs_stopall_nb(long server);
/* Same as `cs_stop_nb' for all requests going to the server `server' or to
   all servers, if `server' is -1. */

void cs_continue_nb(CS_REQTYPE *req);
/* Continues the request `req'. This means, that the `stopped' flag is
   unset and - if the corresponding server has no current request - the
   first request which is not yet finished and is not stopped is begun,
   which can be `req'. If `cs_automatic_io' is set, `cs_do_io' is called
   afterwards. */

void cs_continueall_nb(long server);
/* Same as `cs_continue_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */

void cs_repeat_nb(CS_REQTYPE *req);
/* The request `req' is repeated. This means nothing if the request is not
   yet begun (no I/O yet). If it is already completed it will be marked as
   `not yet started' and will be sent again to the server. The previous
   answer will be overwritten. Nothing happens to a request which has
   already started to do I/O but is not yet finished. The user has to make
   sure, that this is handled! */

void cs_repeatall_nb(long server);
/* Same as `cs_repeat_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */

void cs_cancel_nb(CS_REQTYPE *req);
/* The request `req' is cancelled. This has the following consequences: A
   request which has not yet started to do I/O is deleted by
   `cs_forget'. If I/O has already started, the flag `cancelled' is set,
   which means, that the request is deleted by `cs_forget' once it is
   completed, such that the protocol with the server is not disturbed. A
   request which is already completed is deleted by `cs_forget'. Note
   that the buffers for this request are freed, when the request is
   deleted! */

void cs_cancelall_nb(long server);
/* Same as `cs_repeat_nb' for all requests going to the server `server'
   or to all servers, if `server' is -1. */


/* For the server: */

void cs_start_server(CS_WORKER workers[],unsigned short int port);
/* Starts the main loop of the server program. All data is
   initialized. This main loop handles all the connections, input and
   output. Once a complete request is received, the corresponding
   worker function is called, according to the magic value. This has
   to allocate the answer buffer and return the result in there. It
   does not have to worry about sending it. The buffer will be freed
   automatically, when all data is transferred. This function
   terminates never. */

