/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * include/dcutil.h   headers for the helper functions (hashes/lists).
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This header file describes the data structures and functions for hashes
 * and lists in the helper functions package "dcutil".
 * 
 * $Id: dcutil.h,v 1.3 2000/04/28 09:44:19 neunhoef Exp $
 *
 */

/* read only once */
#ifndef DCUTIL_H
#define DCUTIL_H 1


#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#define HTFAIL LONG_MAX

/* type for lists of long */
typedef struct {
  long len;    /* number of entries */
  long maxlen; /* allocated space for entries */
  long * list; /* the actual list */
} LongList;

/* Things to work with lists of pointers: */
typedef struct {
  long len;      /* number of pointers in this list */
  long maxlen;   /* length of the memory area */
  void **list;   /* pointer to an array of pointers */
} PtrList;

/* type for hash tables with value entries of type long */
typedef struct {
  long len;    /* allocated length = max. nr of entries + 1 */
  long nr;     /* number of entries */
  void **els;   /* the elements */
  long *vals;  /* corresponding values */
  long (*hfun)(void *, size_t);  /* the hash function (pointer to element and
				    length of data for element */
  int (*cmp)(void *, void *, size_t); /* compare function of two elements,
					 given by void pointers and length */
} HashType;


/* implemented in dcutil.c */

HashType *NewHT( long, long (*hfun)(void *, size_t),
		 int (*cmp)(void *, void *, size_t));

HashType *NewHTNoVals( long, long (*hfun)(void *, size_t),
		       int (*cmp)(void *, void *, size_t));

long ValueHT(HashType *, void *, size_t);

long AddHT(HashType *, void *, size_t, long);

void ClearHT(HashType *); 

LongList *NewLongList(long); 

long AddLongList(LongList *, long);

long AppendZerosLongList(LongList *, long);

PtrList *NewPtrList(long slen);
/* Allocates a new PtrList for at most `slen' pointers. The length is
   set to 0. Returns pointer to a PtrList struct. Terminates, if
   something goes wrong. */

long AddPtrList(PtrList *l, void *v);
/* Adds a pointer at the end of the PtrList *l. If this list has not
   enough room, it is enlarged by a factor of 3/2 by realloc. If this
   does not work, the program is terminated. The new length is
   returned. */

long AppendZerosPtrList(PtrList *l, long len);
/* Ensures that the list l has enough room for at least `len'
   pointers. This is accomplished by `realloc'ing the buffer and
   appending zero pointers. If something goes wrong, a negative number 
   is returned and 0 otherwise. */

void dcreadline(char *b,size_t l,FILE *f);
/* Reads a line from file f into buffer b which has space for l
   bytes. Ignores everything after the first # character and ignores
   empty lines consisting only of whitespace. */

ssize_t dcreadfile(FILE *file,char **buffer);
/* Reads a text file into memory. The file is read completely in one block.
   Then in every line everything from the first # sign on is deleted. The
   memory needed is fetched via `malloc'. A pointer on the remaining text
   is stored in *buffer. After removal of comments the memory is realloced
   to the actuall length. The number of bytes in the buffer (can be 0) is 
   returned or -1 in case of an error. The idea is to use `dcfindtoken'
   and `sscanf' to grab the values. */

char *dcfindtoken(char **buf,char *bufend);
/* First advances *buf over all whitespace (spaces, tabs, newlines), then
   stores this position, then advances *buf further up to the next whitespace
   character, places a zero character, and advances *buf one more character.
   returns stored position. *buf is never advanced to a value greater than
   bufend. Assumes that it is possible (and safe) to place a zero byte at
   the address bufend points to. Each call to `dcfindtoken' returns a pointer 
   to a zero-terminated string containing exactly one token. */

#endif  /* reading this file */
