/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * include/grease.h   Headers for greasing layer between MeatAxe and app.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This header file describes the data structures and functions for the
 * intermediate "greasing" layer between the C-MeatAxe and applications
 * using fast vector-matrix multiplication over small finite fields.
 * 
 * $Id: grease.h,v 1.6 2000/04/28 09:44:19 neunhoef Exp $
 *
 */

/* This file implements greasing methods for vector matrix multiplication
   using the MeatAxe library for ground fields with less than 257 elements. */

#include "dcutil.h"
#include "meataxe.h"

/* type for greased matrices: */

typedef struct {
  long ***tabs; /* tables for different remainders 
	 	   of byte numbers mod grrows */
  int *nrvals;  /* number of values produced by each table */
  int nrtabs;   /* number of tables used */
} EXTRTAB;

typedef struct {
  long noc;         /* number of columns of matrix */
  long nor;         /* number of rows of matrix */
  long fl;          /* number of elements in the ground field */
  long grrows;      /* number of rows which are greased */
  long vinblock;    /* number of vectors in a block = fl^grrows */
  EXTRTAB *extrtab; /* some tables to speed up extraction */
  int MPB;          /* number of marks per byte */
  PTR mat;          /* Pointer to a memory area with the precalc data */
} GreasedMatType;

/* extrtab: is used when the grease value `grrows' is a divisor of the number
   of field elements per byte and we are not in GF(2) with grrows == 1. 
   Then we can grab out bytes, look into extrtab and get one to four values. 
   We assume that a long has at least 4 Bytes. Endianess plays a role here! */

/* The precalculated data looks as follows: for each "grease block", that is
   a block of `grrows' rows in the original matrix, we store first a
   struct GreaseBlockHeader, which is a pointer to the data for the next
   grease block, then a long for the offset and a long for the length
   of the data for each row, as the function zaddrowpartial needs it
   (in long words for characteristic 2 and in bytes otherwise). After
   that follow all the possible linear combinations (except the zero
   linear combination!), stored in compact form. After the last grease
   block the remaining rows follow in normal form. */

typedef struct GreaseBlockHeader {
  struct GreaseBlockHeader *next;  /* pointer to next grease block */
  long first;        /* number of long words skipped */
  long len;          /* number of valid long words */
} GREASEBLOCKHEADER;

GreasedMatType *GreaseMeatAxeMat(matrix_t *M,long grrows);
/* Does the precalculation for greasing vector matrix multiplication for
   the matrix M. Greasing is done with bunches of `rows' rows, which means, 
   that all linear combinations of those rows are calculated. Returns NULL,
   if there is not enough memory available. M is freed. It is allowed to 
   set `grrows' to 0, then no greasing is used and the data is not copied. */

void GreaseMapRow(PTR v,GreasedMatType *M,PTR w);
/* Calculates the matrix product of the vector `v' with the matrix `M'. The 
   result is stored in `w'. The length of `v' must coincide with the number 
   of rows of `M'. The result `w' is as long as the length of a row of
   `M'. Sets field and row length correctly! This works also, if `grrows'
   is zero, which means, that greasing is switched off! */


