/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * mtx_vec_U_quo/custom.c   Implementation of operation matrices on vectors
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * In this file the spezialization to matrices acting on vectors
 * by right multiplication using the C-MeatAxe is implemented. The
 * "U-quotient"-trick is used to handle vector recognition by
 * enumerating orbits "by U-orbits". 
 * 
 *  $Id: custom.c,v 1.7 2000/04/28 12:27:14 neunhoef Exp $
 */


/* include files for the application */
#include "grease.h"
#include "custom.h"

/* define's here */

/* extra variables */
#ifdef DC_ISCLIENT

static int fieldsize;   /* number of elements in ground field */
static int dimension;   /* dimension of vector space of matrix representation */
static int codimension; /* codimension of U-invariant subspace */

static PtrList *uels;   /* all the elements of U */
static LongList *precalc;  /* For every possible vector in the
			      quotient space V/W we store either a
			      nonnegative number which is the number
			      of an element of U to apply to find the
			      minimal element or the negative of a
			      number of a U orbit in V/W */
static PtrList *stabs;  /* For every U orbit in V/W we store the
			   stabilizer as a list of numbers of elements 
			   of U, so every Ptr is a pointer to a
			   LongList */
static HashType *hU;    /* A hash table for the enumeration of U-orbits */
static long ksize;      /* number of elements in K */
static int nu;          /* number of generators of U */
static matrix_t **ugen; /* the generators */
static PtrList *uelsf;  /* all the elements of U in the representation 
			   on the factor */
static LongList *invtab;   /* Table of inverses for the elements of U */
static int grrowsG;     /* number of rows to grease for matrices of G */
static int grrowsK;     /* number of rows to grease for matrices of K */
static int grrowsU;     /* number of rows to grease for matrices of U */
#endif
/* ifdef DC_ISCLIENT */


/* The following functions are used only in the client: */

#ifdef DC_ISCLIENT
    
/* internal subroutines: */

static int MatCmp(matrix_t *el1,matrix_t *el2)
/* A comparison function for matrices of U. Assumes the MeatAxe
   library is already initialized. Returns 0 if the matrices are identical 
   and 1 otherwise. */
{
  int i;
  PTR p1,p2;

  if (el1->noc != el2->noc || el1->nor != el2->noc || el1->fl != el2->fl)
    return 1;
  zsetfield(el1->fl);
  zsetlen(el1->noc);
  for (i = 0,p1 = el1->d,p2 = el2->d;i < el1->nor;i++) {
    if (zcmprow(p1,p2)) return 1;
    zadvance(&p1,1L);
    zadvance(&p2,1L);
  }
    
  return 0;
}

static unsigned long NumberFFVector(PTR v,int fl,int dim)
/* Implements a numbering of all the vectors in the vector space of
   dimension `dim' over the field `fl'. This function is the inverse
   to `FFVectorNumber'. */
{
  int i;
  unsigned long res;

  res = 0;
  for (i = 1;i <= dim;i++) 
    res = res * fl + zftoi(zextract(v,i));
  return res;
}

static unsigned long NumberFFVectorOffset(PTR v,int fl,int dim,int offset)
/* Does the same as `NumberFFVector' but does not start at the beginning
   of the vector but at the index `offset' for `dim' entries. */
{
  int i;
  unsigned long res;

  res = 0;
  for (i = offset;i < offset+dim;i++) 
    res = res * fl + zftoi(zextract(v,i));
  return res;
}

static void FFVectorNumber(PTR v,unsigned long nr,int fl,int dim)
/* Implements a numbering of all the vectors in the vector space of
   dimension `dim' over the field `fl'. This function is the inverse
   to `NumberFFVector'. */
{
  int i;
  for (i = dim;i > 0;i--) {
    zinsert(v,i,zitof( nr % (unsigned long) fl ));
    nr /= (unsigned long) fl;
  }
}

static void enumerateU(void)
/* This function takes as input the generators of the group U as matrices
   (given in the global variables nu and ugen). `dimension' is used
   for the number of rows of the matrices and it is assumed that the
   MeatAxe library is already initialized and the field and rowsize is
   set. The result is stored in the global variable uels. */
{
  int i,j,k;
  matrix_t *el;  /* space for one matrix: */

  printf("Enumerating U ... "); fflush(stdout);

  uels = NewPtrList(1000L);

  /* First we want to have the identity matrix: */
  AddPtrList(uels,MatId(fieldsize,dimension));

  /* Add the generators into the list of elements: */
  for (i = 0;i < nu;i++)
    AddPtrList(uels,ugen[i]);

  /* Now do the orbit algorithm: */
  for (i = 0;i < uels->len;i++)
    for (j = 0;j < nu;j++) {
      el = MatDup(uels->list[i]);
      el = MatMul(el,ugen[j]);
      for (k = 0;k < uels->len;k++)
	if (!MatCmp(uels->list[k],el))    /* Already there! */
	  break;
      if (k >= uels->len)   /* not yet known */
	AddPtrList(uels,el);
      else
	MatFree(el);
    }

  printf("found %ld elements in the group U\n",uels->len);
  fflush(stdout);
}

static void classifyorbits(void)
/* Classifies all orbits of U on V/W. Uses the data in the global
   variables `uelsf', `invtab', `fieldsize', and `codimension' and
   calculates the values of `precalc' and `stabs'. We run through all
   possible vectors in V/W beginning from the "smaller" ones. We apply
   all elements of U and find in this way all the vectors in the same
   orbit as well all elements of the point stabilizer of the minimal
   vector in this orbit. */
{
  long i, j;
  long numbervalues;
  PTR v,w;
  long nr;
  LongList *newstab;
  long numbervaluesfound;

  printf("Doing precalculation in quotient space ...\n"); fflush(stdout);

  /* we calculate in the quotient here: */
  zsetfield(fieldsize);
  zsetlen(codimension);

  v = zalloc(1L);  /* room for a vector */
  w = zalloc(1L);  /* root for another one */

  /* number of all possible values: */
  numbervalues = fieldsize;
  for (i = 1;i < codimension;i++) numbervalues *= fieldsize;

  /* We need a new empty list: */
  precalc = NewLongList(numbervalues);
  AppendZerosLongList(precalc,numbervalues);
  /* note that the element of U with number 0 is the identity, so 0
     can be used to indicate "not yet found" in this routine! */

  stabs = NewPtrList(1000L);   /* List of stabilizers */
  AppendZerosPtrList(stabs,1); /* Start numbering at 1 for positive values */

  numbervaluesfound = 0;  /* total number of processed values */
  for (i = 0;i < numbervalues;  /* do not advance i here! */) {
    /* we assume to stand on a "not yet found" element here */
    /* first we begin a new U-orbit: */
    /* note that we assume most U-orbits to be regular */
    newstab = NewLongList(4);
    /* note that we do not store the identity with each stabilizer! */
    precalc->list[i] = -stabs->len;  /* the number of the U-orbit */
    numbervaluesfound++;
    AddPtrList(stabs,newstab);
    FFVectorNumber(v,i,fieldsize,codimension);

    /* now apply all elements of U to the vector in question: */
    for (j = 1  /* skip the identity */ ;j < uelsf->len;j++) {
      zmaprow(v,((matrix_t *)uelsf->list[j])->d,codimension,w);
      nr = NumberFFVector(w,fieldsize,codimension);
      if (precalc->list[nr] == 0)   { /* we do not know this vector */
	precalc->list[nr] = invtab->list[j];  /* this is needed to normalize */
	numbervaluesfound++;
      } else if (!zcmprow(v,w))  /* a new stabilizer element! */
	AddLongList(newstab,j);
    }
    if ((stabs->len-1) % 1000 == 0) {
      printf("Found %ld U-orbits, total vectors: %ld, average len: %.2f.\n",
	     stabs->len-1,numbervaluesfound,
	     (double) numbervaluesfound / (stabs->len-1));
      fflush(stdout);
    }

    /* find the next "not yet known" vector: */
    do { i++; } while (i < numbervalues && precalc->list[i] != 0);
  }

  printf("done.\n");
  printf("Found %ld U-orbits with an average length of %.2f vectors.\n",
	 stabs->len-1,(double) numbervalues / (stabs->len-1));
  fflush(stdout);
  free(v);
}

/* Some of the declarations of functions below may be  substituted by 
   macros for some applications */


/* void DC_CLIENT_INIT(FILE *input) 
   this function must set at least the following global variables for the
   client, see the beginning of `dcclientgeneric.c':
   
long dc_length_pt;      length of point in orbit in bytes   
long dc_ng;             number of elements to condense (in startup number
                        of given generators of G   
void **ggen;            list of dc_ng void pointers to generators of G   
long dc_nk;             number of given generators of K   
void **kgen;            list of dc_nk void pointers to generators of K   
void *start;            pointer to start point
long htklen;            length of hash table for K-orbits   
long maxminpart;        maximal number of minimal elements in a part the 
			partitions of a K-orbit
long maxpart;           maximal number of elements in a part
long maxnrpartskorb;    maximal number of parts in a K-orbit
long maxextra;          maximal extra number of minimal representatives per
			K-orbit 
long dc_max_fact_req;   this times nr of hash servers requests to hash 
                        servers are produced at most in `CondMatRows'   

Furthermore it should set up all global data needed by the other functions
defined below.

*/

void DC_CLIENT_INIT(FILE *input)
{
  long i, j;    /* counters */
  char *buf;    /* to read the input file */
  ssize_t len;  /* length of input */
  char *bufend; /* pointer behind end of buf */
  char *p;      /* to walk through the file */
  long fl;      /* ground field */
  long nor;     /* number of rows in matrix */
  long dim;     /* number of columns in matrix */
  matrix_t *m;  /* pointer to a matrix */
  FILE *f;      /* to read start vector */

  /* first initialize the MeatAxe library: */
  if (MtxInit() < 0) {
    fprintf(stderr,"Could not initialize MeatAxe library!\n");
    exit(1);
  }

  /* read the input file: */
  len = dcreadfile(input,&buf);
  if (len < 0) {
    fprintf(stderr,"Could not read input file.\n");
    exit(1);
  }
  bufend = buf + len;
  p = dcfindtoken(&buf,bufend);  /* fetch the first token */

  /* number of elements to condense: */
  sscanf(p, "%ld", &dc_ng); p = dcfindtoken(&buf,bufend);
  if (dc_ng <= 0) {
    fprintf(stderr,"No elements to condense!\n");
    exit(1);
  }
  
  /* elements of G to condense: */
  ggen = (void*)malloc(dc_ng*sizeof(void*));
  for (i=0; i<dc_ng; i++) {
    printf("Reading matrix %s ...",p); fflush(stdout);
    ggen[i] = MatLoad(p);
    if (ggen[i] == NULL) {
      fprintf(stderr, "Cannot read matrix (name=%s).\n",p);
      exit(1);
    }
    printf("done\n"); fflush(stdout);
    p = dcfindtoken(&buf,bufend);
  }

  fieldsize = ((matrix_t *)ggen[0])->fl;
  zsetfield(fieldsize);
  printf("Number of elements in ground field: %d\n",fieldsize);
  dimension = ((matrix_t *)ggen[0])->noc;
  zsetlen(dimension);
  printf("Dimension of representation space: %d\n",dimension);
  fflush(stdout);
  /* Now we know the length of a point, which is a vector here: */
  dc_length_pt = zsize(1L);

  /* greasing for matrices for G: */
  sscanf(p, "%d", &grrowsG); p = dcfindtoken(&buf,bufend);

  printf("Doing precalculation for greasing (%d) for G ... ",grrowsG); 
  fflush(stdout);
  for (i = 0;i < dc_ng; i++) {
    ggen[i] = GreaseMeatAxeMat(ggen[i],grrowsG);
    if (ggen[i] == NULL) {
      fprintf(stderr,"Not enough memory for greasing!\n");
      exit(1);
    }
    printf("*"); fflush(stdout);
  }
  printf(" done\n"); fflush(stdout);

  /* number of generators of K: */
  sscanf(p, "%ld", &dc_nk); p = dcfindtoken(&buf,bufend);
  /* generators of K */
  kgen = (void*)malloc(dc_nk*sizeof(void*));
  for (i=0; i<dc_nk; i++) {
    printf("Reading matrix %s ...",p); fflush(stdout);
    kgen[i] = MatLoad(p);
    if (kgen[i] == NULL) {
      fprintf(stderr, "Cannot read matrix (name=%s).\n",p);
      exit(1);
    }
    printf("done\n"); fflush(stdout);
    p = dcfindtoken(&buf,bufend);
  }

  /* greasing for matrices for K: */
  sscanf(p, "%d", &grrowsK); p = dcfindtoken(&buf,bufend);

  printf("Doing precalculation for greasing (%d) for K ... ",grrowsK); 
  fflush(stdout);
  for (i = 0;i < dc_nk; i++) {
    kgen[i] = GreaseMeatAxeMat(kgen[i],grrowsK);
    if (kgen[i] == NULL) {
      fprintf(stderr,"Not enough memory for greasing!\n");
      exit(1);
    }
    printf("*"); fflush(stdout);
  }
  printf(" done\n"); fflush(stdout);

  /* number of generators of U: */
  sscanf(p, "%d", &nu); p = dcfindtoken(&buf,bufend);
  /* generators of Z: */
  ugen = malloc(nu*sizeof(matrix_t *));
  for (i=0; i<nu; i++) {
    printf("Reading matrix %s ...",p); fflush(stdout);
    ugen[i] = MatLoad(p);
    if (ugen[i] == NULL) {
      fprintf(stderr, "Cannot read matrix (name=%s).\n",p);
      exit(1);
    }
    printf("done\n"); fflush(stdout);
    p = dcfindtoken(&buf,bufend);
  }

  /* greasing for matrices for U: */
  sscanf(p, "%d", &grrowsU); p = dcfindtoken(&buf,bufend);

  /* Read the codimenstion of the U-invariant subspace: */
  sscanf(p,"%d",&codimension); p = dcfindtoken(&buf,bufend);

  /* start point: */
  start = zalloc(1L);
  if ((f = zreadhdr(p, &fl, &nor, &dim)) == NULL) {
    fprintf(stderr, "Cannot read vector (name=%s).\n",p);
    exit(1);
  }
  if (fieldsize != fl || nor != 1 || dimension != dim) {
    fprintf(stderr, "Field or size of vector not correct (name=%s).\n",p);
    exit(1);
  }
  zreadvec(f, start, 1);
  fclose(f);
  p = dcfindtoken(&buf,bufend);

  /* number of elements in K: */
  sscanf(p, "%ld", &ksize); p = dcfindtoken(&buf,bufend);

  /* length of hash tables for K-orbits: */
  sscanf(p, "%ld", &htklen); p = dcfindtoken(&buf,bufend);

  /* this times nr of hash servers requests to hash servers are 
     produced at most in `CondMatRows' */   
  sscanf(p, "%ld", &dc_max_fact_req); p = dcfindtoken(&buf,bufend);

  /* now we have to do some precalculations: */
  
  /* first we do an orbit algorithm for the elements of U: */
  enumerateU();

  /* now we extract the matrices for the operation of U on the factor: */
  uelsf = NewPtrList(uels->len);
  for (i = 0;i < uels->len;i++)
    AddPtrList(uelsf,MatCut(uels->list[i],
			    dimension-codimension+1,dimension-codimension+1,
			    codimension,codimension));

  /* we compute a table of the inverses of all group elements of U: */
  printf("Computing inverses of elements of U ... "); fflush(stdout);
  invtab = NewLongList(uels->len);
  for (i = 0;i < uels->len;i++) {
    m = MatInv(uels->list[i]);
    for (j = 0;j < uels->len;j++) {
      if (!MatCmp(m,uels->list[j])) {
	invtab->list[i] = j;
	break;
      }
    }
    MatFree(m);
    if (j >= uels->len) {
      fprintf(stderr,"Did not find inverse of some element of U!\n");
      exit(1);
    }
  }
  printf("done.\n"); fflush(stdout);

  /* Now we completely classify all orbits of U in its operation on
     the factor of the vector space V by the U-invariant subspace W. */
  classifyorbits();

  /* maximal number of minimal elements in part (=U-orbit) of K-orbit: */
  maxminpart = uels->len;  /* for the value of zero in the quotient */
  /* maximal number of elements in a part (=U-orbit): */
  maxpart = uels->len;
  /* maximal number of parts in a K-orbit: */
  maxnrpartskorb = ksize / uels->len;  /* this is a lemma: each K-orbit is
       a disjoint union of U-orbits, a transversal of U in K is enough to
       find all those U-orbits. */
  /* maximal extra number of minimal representatives per K-orbit: */ 
  maxextra = maxnrpartskorb * maxminpart;
  /* this is the worst case, but it could happen! */

  printf("Doing precalculation for greasing (%d) for U ... ",grrowsU); 
  fflush(stdout);
  for (i = 0;i < uels->len; i++) {
    uels->list[i] = GreaseMeatAxeMat(uels->list[i],grrowsU);
    if (uels->list[i] == NULL) {
      fprintf(stderr,"Not enough memory for greasing!\n");
      exit(1);
    }
    printf("*"); fflush(stdout);
  }
  printf(" done\n"); fflush(stdout);

  /* prepare the MeatAxe library: */
  zsetfield(fieldsize);
  zsetlen(dimension);

  /* Just initialize this once and for all: */
  hU = NewHTNoVals(2*uels->len,DC_HASHFUN,DC_CMP_PTS);
}

/* void DC_OPERATION(void* pt, void *g, void *img)
   This should compute the image of point pt under group element g and 
   store the resulting point in img. Points and group elements are referred 
   to by void pointers, img must point to a large enough part of the memory 
   to hold the result.   
*/
void DC_OPERATION(void* pt, void *g, void *img)
{
  PTR v = (PTR) pt;
  PTR w = (PTR) img;
  GreasedMatType *m = g;

  GreaseMapRow(v,m,w);
}

/* long DC_HASHSERVFUN(void *);
   The hash server function, it gets a point as one argument and has
   to return the number of a hash server in the range from
   0..#hashservers-1: */
long DC_HASHSERVFUN(void *pt)
{ 
  /* just one hash server: */
  if (nrhashserv == 1) return 0;
  else return (*((unsigned long *)pt) / 100) % nrhashserv;
}

/* void DC_FINDMINIMAL(void *pt1, void *pt2);
   This function is called with a point pt1 and has to store one 
   minimal point in the same part of the K-orbit into pt2. 
   pt2 must point to a large enough part of the memory.
*/
void DC_FINDMINIMAL(void *pt1, void *pt2)
{
  PTR v1 = (PTR) pt1;
  PTR v2 = (PTR) pt2;
  long nr = NumberFFVectorOffset(v1,fieldsize,codimension,
				 dimension-codimension+1);
  if (precalc->list[nr] < 0)   /* this is a minimal vector! */
    zmoverow(v2,v1);
  else
    GreaseMapRow(v1,uels->list[precalc->list[nr]],v2);
}

/* void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   minimal points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found minimal points.
*/
void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr)
{
  PTR p = (PTR) ptr;
  PTR q;
  PTR v;         /* space for one vector */
  long i,j;
  long orbnr;    /* number of the U-orbit */
  LongList *st;  /* The stabilizer in question */
  long nrmins;   /* current number of minimal vectors found */

  v = zalloc(1L);
  DC_FINDMINIMAL(pt1,ptr);
  nrmins = 1;
  zadvance(&p,1L);
  orbnr = NumberFFVectorOffset(ptr,fieldsize,codimension,
			       dimension-codimension+1);
  orbnr = -precalc->list[orbnr]; 
    /* orbnr will now be the number of an U-orbit */
  st = (LongList *)(stabs->list[orbnr]);  /* the stabilizer */
  for (i = 0;i < st->len;i++) {  /* the identity is *not* stored! */
    GreaseMapRow((PTR) ptr,uels->list[st->list[i]],v);
    /* Do we know this vector already? */
    for (j = 0,q = ptr;j < nrmins;j++) {
      if (zcmprow(q,v))  /* not yet found */
	zadvance(&q,1L);
      else
	break;
    }
    if (j >= nrmins) {   /* not found! */
      zmoverow(p,v);   /* copy it */
      zadvance(&p,1);
      nrmins++;
    }
  }
  free(v);
  *nr = nrmins;
}

/* void DC_FINDPART(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found  points.
*/
void DC_FINDPART(void *pt1, void *ptr, long *nr)
{
  PTR p;
  long i, j;
  PTR v;

  v = zalloc(1);
  ClearHT(hU);

  p = (PTR) ptr;
  zmoverow(p,pt1);
  AddHT(hU,p,zsize(1),0);
  zadvance(&p,1);
  j = 1;

  for (i = 1; i < uels->len; i++) {
    GreaseMapRow((PTR) ptr,uels->list[i],v);
    if (ValueHT(hU, v, zsize(1)) == HTFAIL) {
      zmoverow(p,v);
      AddHT(hU, p, zsize(1), 0 );
      zadvance(&p,1);
      j++;
    }
  }
  free(v);
  *nr = j;
}

#endif
/* ifdef DC_ISCLIENT */

/* the following functions are used by the client and the hashserver: */

/* int DC_CMP_PTS(void* pt1, void* pt2, size_t len) 
   This is for comparing two points (of data length `len'), returns 
   0 in case of equality. 
*/
int DC_CMP_PTS(void *pt1, void *pt2, size_t len)
{
  len = len;  /* unused */
  return zcmprow((PTR) pt1, (PTR) pt2);
}

/* long DC_HASHFUN(void *pt, size_t len);
   The hash function, it gets a point as one argument and the length of 
   its data, it has to return a long number.
*/
long DC_HASHFUN(void *pt, size_t len)
{
  int i;
  long res;
  unsigned char *p;

  len = len; /* unused */

  for (i = 0,p = pt,res = 0;(size_t) i < zrowsize_io;i++,p++)
    res = res * 137 + (unsigned long) *p;

  return res;
}

/* The following functions are used only in the hash server: */

#ifndef DC_ISCLIENT

/* void DC_HASHSERVINIT(LongList *l);
   This function  is called with some data in a long list to configure 
   some global variables in a hash server. */
void DC_HASHSERVINIT(LongList *l)
{
  MtxInit();
  /* we get only the field size and the dimension: */
  zsetfield(l->list[1]);
  zsetlen(l->list[2]);
  dc_length_pt = zsize(1);
}

#endif
/* ifndef DC_ISCLIENT */


