/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * mtxgeneric/grease.c   Implementation of intermediate layer for greasing.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This file contains code for an intermediate layer of code between
 * the C-MeatAxe and a program using vector matrix multiplication.
 * It implements R. Parkers "greasing" techniques for this case.
 * 
 * $Id: grease.c,v 1.14 2000/04/28 13:15:22 neunhoef Exp $ 
 *
 */

/* This file implements greasing methods for vector matrix multiplication
   using the MeatAxe library for ground fields with less than 257 elements. */

#include "grease.h"

static EXTRTAB *BuildExtractionTable(int fl,int grrows)
/* Builds the tables used for the extraction of grease bits. 
   It can build tables for the following field/grease combinations:
   field: 2-256, grrows: 1-16 */
{
  int MPB;     /* marks per byte */
  EXTRTAB *t;  /* the result */
  unsigned char c[4];   /* buffer for calculations */
  long flpowtab[17];    /* powers of fl */
  long *w;     /* to write into tables */
  long i,j,k,l;  /* counters */
  int extraspace;
  int nrvals;
  int restbits;

  t = malloc(sizeof(EXTRTAB));
  if (t == NULL) {
    fprintf(stderr,"Not enough memory for extraction table!\n");
    exit(2);
  }

  zsetfield(fl);
  zsetlen(sizeof(long)*2*8*5*3);
  /* Note that sizeof(long)*2*8*5*3 will be divisable by MPB*sizeof(long)*2,
     so a vector of this length will completely use all zsize(..) bytes.
     We need the *2 because of possible MMX modifications. */
  MPB = sizeof(long) * 2 * 8*5*3 / zsize(1);

  /* now we need the lcm of MPB and grrows: */
  for (l = 1;(l * MPB) % grrows != 0;l++);
  /* now there are l different types of bytes, so we need l tables: */
  t->nrtabs = l;
  t->tabs = malloc(sizeof(long **) * l);
  t->nrvals = malloc(sizeof(int) * l);
  if (t->tabs == NULL || t->nrvals == NULL) {
    fprintf(stderr,"Not enough memory for extraction table!\n");
    exit(2);
  }
  
  /* calculate flpowtab: */
  for (l = 0,j = 1;l <= 16;l++,j *= fl) 
    flpowtab[l] = j;
 
  for (l = 0;l < t->nrtabs;l++) {   /* All tables */
    t->tabs[l] = calloc(256,sizeof(long *));
    if (t->tabs[l] == NULL) {
      fprintf(stderr,"Not enough memory for extraction table!\n");
      exit(2);
    }

    /* Calculate the number of values completed in this byte: */
    nrvals = 0;   /* we count the number of values */
    if ((l*MPB) % grrows > 0) {
      restbits = grrows - (l*MPB) % grrows;
      if (restbits <= MPB)   /* one value to complete: */
	nrvals++;
    } else
      restbits = 0;
    nrvals += (MPB - restbits) / grrows;
    t->nrvals[l] = nrvals;
    if (((l+1)*MPB) % grrows > 0)   /* one value only begun, we need space! */
      extraspace = 1;
    else
      extraspace = 0;

    /* Now we go through all possibilities of this byte: */
    for (i = 0;i < flpowtab[MPB];i++) {
      *c = 0;
      for (j = 1,k = i;j <= MPB;j++) {
	zinsert((PTR) c,j,zitof(k % fl));
	k /= fl;
      }
      t->tabs[l][*c] = calloc(nrvals+extraspace,sizeof(long));
      if (t->tabs[l][*c] == NULL) {
	fprintf(stderr,"Not enough memory for extraction table!\n");
	exit(3);
      }

      /* now distribute the different values in i to the table: */
      w = t->tabs[l][*c];
      k = i;

      /* first (perhaps) a value that already begun in the previous byte: */
      if (restbits > 0) {  /* one value to complete: */
	if (restbits <= MPB) {
	  /* is completed within this byte! */
	  *w++ = flpowtab[(l*MPB) % grrows] *
     	         (k % flpowtab[restbits]);
	  k /= flpowtab[restbits];
	} else {
	  /* has already started and does not end within this byte: */
	  *w++ = flpowtab[(l*MPB) % grrows] * k;
	  continue;  /* to avoid the code for the last value */
	}
      }
      /* Now all the values fully within the byte: */
      for (j = (MPB - restbits) / grrows;j>0;j--) {
	*w++ = k % flpowtab[grrows];
	k /= flpowtab[grrows];
      }
      /* Now (perhaps) a value that is only begun: */
      if (((l+1)*MPB) % grrows > 0) /* one additional part-value */
	*w++ = k;
    }
  }
  return t;
}

static EXTRTAB *extrtabcache[257][17];
static extrtabcacheinitialized = 0;

static EXTRTAB *FetchExtractionTable(int fl,int grrows)
/* Fetches a table either from cache or uses BuildExtractionTable to
   build it. */
{
  int i,j;

  if (!extrtabcacheinitialized) {
    for (i = 2;i <= 256;i++)
      for (j = 1;j <= 16;j++)
	extrtabcache[i][j] = NULL;
    extrtabcacheinitialized = 1;
  }

  if (fl < 2 || fl > 256) return NULL;
  if (grrows < 1 || grrows > 16) return NULL;

  if (extrtabcache[fl][grrows] == NULL)
    extrtabcache[fl][grrows] = BuildExtractionTable(fl,grrows);

  return extrtabcache[fl][grrows];
}

GreasedMatType *GreaseMeatAxeMat(matrix_t *M,long grrows)
/* Does the precalculation for greasing vector matrix multiplication for
   the matrix M. Greasing is done with bunches of `rows' rows, which means, 
   that all linear combinations of those rows are calculated. Returns NULL,
   if there is not enough memory available. M is freed. It is allowed to 
   set `grrows' to 0, then no greasing is used and the data is not copied. */
{
  long i,j,k,l; /* counters */
  long rows;    /* number of rows written so far in current part */
  PTR p;        /* to write vectors */
  PTR q;        /* to read vectors from the original matrix */
  PTR r;        /* to read vectors from previous results */
  PTR v;        /* space for one vector */
  long nrvecs;  /* number of vectors in greased table */
  GreasedMatType *res;  /* the result */
  PTR buffer;   /* One buffer for a grease block */
  long first;   /* first value != 0 */
  long last;    /* last value != 0 */
  long *x;      /* to search zeros */
  long LPR;     /* longs per row */
  long len;     /* number of longs */
  PTR *lastp;   /* to make the link */
  GREASEBLOCKHEADER *gbh;

  res = malloc(sizeof(GreasedMatType));
  if (res == NULL) return NULL;
  res->fl = M->fl;
  res->noc = M->noc;
  res->nor = M->nor;
  res->grrows = grrows;

  zsetfield(M->fl);
  zsetlen(M->nor);

  /* special case of greasing switched off: */
  if (grrows == 0) {
    res->vinblock = 1;
    res->mat = M->d;
    res->extrtab = NULL;
    free(M);  /* note that we do not use MatFree in this case! */
    return res;
  }

  nrvecs = 1;
  for (i = grrows;i > 0;i--) nrvecs *= M->fl; /* This calculates fl^grrows */
  res->vinblock = --nrvecs;  /* the null vector is not stored! */

  buffer = calloc(nrvecs,zsize(1));
  if (buffer == NULL) {
    free(res);
    return NULL;
  }

  lastp = &(res->mat);  /* this will be initialized later! */

  v = zalloc(1);

  /* Now we calculate all linear combinations necessary: */
  q = M->d;
  for (i = M->nor / grrows;i > 0;i--) {  /* all greasing block */
    p = buffer;
    rows = 0;  /* this is 1 - 1 = fl^0 - 1 */
    LPR = zsize(1) / sizeof(long);
    first = LPR;
    last = 0;

    /* now for all vectors in the block: */
    for (j = grrows;j > 0;j--) {   /* all vectors in block */
      
      for (x = (long *) q,k = 0;k < LPR && *x == 0;k++,x++);
      if (k < first) first = k;
      for (x = ((long *) q) + LPR-1,k = LPR-1;k > 0 && *x == 0;k--,x--);
      if (k > last) last = k;

      for (k = 1;k < M->fl;k++) {  /* all field elements */
	zmoverow(v,q);
	zmulrow(v,zitof(k));
	zmoverow(p,v);   /* copy the new multiple */
	zadvance(&p,1);

	r = buffer;    /* start from the beginning of the current block */
	for (l = rows;l > 0;l--) {   /* for all vectors so far */
	  zmoverow(p,r);
	  zadvance(&r,1);
	  zaddrow(p,v);
	  zadvance(&p,1);
	}
      }
      zadvance(&q,1);   /* take a new row of the original matrix */
      rows = (rows+1)*M->fl - 1;  /* the null vector is not there */
    }
    /* now `first' and `last' are up to date with respect to all original
       rows in the current grease block, we can compress the data: */
    len = last+1-first;
#ifdef ASM_MMX
    /* make `len' an even number due to 64bit access: */
    if (len & 1) {
      len++;
      if (first+len > LPR)
	first--;
    }
#endif
    gbh = (GREASEBLOCKHEADER *) 
          calloc(nrvecs*len*sizeof(long) + sizeof(GREASEBLOCKHEADER),1);
    *lastp = (PTR) gbh;
    gbh->next = NULL;
    lastp = (PTR *) &(gbh->next);
    gbh->first = first;
    gbh->len = len;
    x = (long *) buffer;
    p = (PTR) ((char *) gbh + sizeof(GREASEBLOCKHEADER));
    for (j = rows;j > 0;j--) {
      x += first;
      memcpy(p,x,len*sizeof(long));
      p = (PTR) ((char *)p + len*sizeof(long));
      x += (LPR - first);
    }
  }

  p = calloc(M->nor % grrows,zsize(1));
  *lastp = p;

  for (i = M->nor % grrows;i > 0;i--) {  /* the rest of the vectors */
    zmoverow(p,q);
    zadvance(&p,1);
    zadvance(&q,1);
  }

  res->extrtab = FetchExtractionTable(M->fl,grrows);
  
  MatFree(M);
  
  free(v);
  free(buffer);

  return res;
}

static long *extractednrs = NULL;
static long extrtablen = 0;

static void ExtractNrs(PTR v,GreasedMatType *M)
/* Builds a table of numbers extracted from the vector v. Stores it into
   extractednrs (global variable, see above). This is malloced anew if
   not long enough! */
{
  unsigned char *p = v;
  register long *q;
  register long curtab = 0;
  register int i = M->nor / M->grrows;

  if (extrtablen < i+16) {   /* some reserve! */
    if (extractednrs) free(extractednrs);
    extractednrs = malloc((i+16) * sizeof(long));
    if (extractednrs == NULL) {
      fprintf(stderr,"Not enough memory for extraction table!\n");
      exit(1);
    }
    extrtablen = i+16;
  }
  q = extractednrs;

  {
    register long y = 0;
    register long *x;
    
    while (i > 0) {
      /* tabnr = 0; */
      x = M->extrtab->tabs[curtab][*p++];
      { 
	register long k = M->extrtab->nrvals[curtab];
	while (k--) {
	  *q++ = y + *x++;
	  y = 0;
	  i--;
	}
      }
      if (++curtab < M->extrtab->nrtabs)
	y += *x;  /* remember last part for "carry" */
      else
	curtab = 0;
    }
  }
}

static void CleverMapRow(PTR v,GreasedMatType *M,PTR w)
/* Calculates the matrix product of the vector `v' with the matrix `M'. The 
   result is stored in `w'. The length of `v' must coincide with the number 
   of rows of `M'. The result `w' is as long as the length of a row of
   `M'. This is called from `GreaseMapRow' in the case where we have
   an extraction table. */
{
  register long *nr;
  register long i;
  long curcol;
  register GREASEBLOCKHEADER *p = (GREASEBLOCKHEADER *) (M->mat);

  ExtractNrs(v,M);   /* extracts numbers into global table */
  zmulrow(w,F_ZERO);  /* Clear result */
  curcol = M->nor / M->grrows;  /* only used later and for i init */
  nr = extractednrs;
  for (i = curcol;i > 0;i--) {  /* for all greasing blocks */
    if (*nr) {  /* not the zero vector */
      register PTR q = (PTR) ((char *) p + sizeof(GREASEBLOCKHEADER));
      zaddrowpartial(w,(PTR) ((char *)q + p->len * sizeof(long) * (*nr-1)),
		     p->first,p->len);
    }
    p = p->next;
    nr++;
  }
  curcol = 1 + curcol * M->grrows;
  { 
    register PTR q = (PTR) p;
    register long zs = zsize(1);

    for (i = M->nor % M->grrows;i > 0;i--) {  /* do the rest */
      zaddmulrow(w,q,zextract(v,curcol++));
      /* zadvance(&q,1); */
      q = (PTR) ((char *) q + zs);
    }
  }
}

void GreaseMapRow(PTR v,GreasedMatType *M,PTR w)
/* Calculates the matrix product of the vector `v' with the matrix `M'. The 
   result is stored in `w'. The length of `v' must coincide with the number 
   of rows of `M'. The result `w' is as long as the length of a row of
   `M'. Sets field and row length correctly! This works also, if `grrows'
   is zero, which means, that greasing is switched off! */
{
  long i,j;     /* Counters */
  long curcol;  /* Current column of `v' */
  long nr;      /* number of vector in the lookup table */
  PTR p;        /* Pointer into lookup table */

  zsetfield(M->fl);
  zsetlen(M->noc);

  if (M->extrtab) {  /* we use our clever extraction! */
    CleverMapRow(v,M,w);
    return;
  }

  if (M->grrows == 0) {  /* actually, we do not grease! */
    zmaprow(v,M->mat,M->nor,w);
    return;
  }

  zmulrow(w,F_ZERO);  /* Clear result */
  curcol = 1;         /* start with the first column of `v' */
  p = M->mat;         /* start at the beginning of the table */

  if (M->fl == 2) {
    register signed char *q = (signed char *) v;
    register signed char buf = *q++;
    register long wmask;   /* used to alter nr */
    for (i = M->nor / M->grrows;i > 0;i--) {  /* for all greasing blocks */
      /* Calculate the number for the first lookup: */
      nr = 0;
      wmask = 1;
      for (j = M->grrows-1;j >= 0;j--) {
	if (buf < 0)  /* a bit is set */
	  nr |= wmask;
	wmask <<= 1;
	if ((curcol++ & 0x7) == 0)
	  buf = *q++;
	else
	  buf <<= 1;
      }
      if (nr) {  /* not the zero vector */
	zadvance(&p,nr-1);   /* now p points to the right linear combination */
	/* Should be: zaddrow(w,p); */
	{
	  register long *a = (long *) w;
	  register long *b = (long *) p;
	  register long k = zsize(1) / sizeof(long);
	  for (;k;k--) *a++ ^= *b++;
	}
	zadvance(&p,M->vinblock - nr + 1);
      } else zadvance(&p,M->vinblock);
    }
    for (i = M->nor % M->grrows;i > 0;i--) {  /* do the rest */
      if (buf < 0) zaddrow(w,p);
      zadvance(&p,1);
      if ((curcol++ & 0x7) == 0)
	buf = *q++;
      else
	buf <<= 1;
    }
  } else {  /* not GF(2) */
    for (i = M->nor / M->grrows;i > 0;i--) {  /* for all greasing blocks */
      /* Calculate the number for the first lookup: */
      nr = 0;
      for (j = M->grrows-1;j >= 0;j--)   /* add j to curcol */
	nr = nr * M->fl + zftoi(zextract(v,curcol+j));
      if (nr) {  /* not the zero vector */
	zadvance(&p,nr-1);   /* now p points to the right linear combination */
	zaddrow(w,p);
	zadvance(&p,M->vinblock - nr + 1);
      } else zadvance(&p,M->vinblock);
      curcol += M->grrows;
    }
    for (i = M->nor % M->grrows;i > 0;i--) {  /* do the rest */
      zaddmulrow(w,p,zextract(v,curcol++));
      zadvance(&p,1);
    }
  }
}
