/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * perms_S21_8841/custom.c   Implementation of operation of S21
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This file contains the spezialization of the operation to the case 
 * of S21 operating in its natural permutation representation. The 
 * condensation subgroup is a 8841 Young subgroup.
 * 
 *  $Id: custom.c,v 1.4 2000/04/28 13:36:18 neunhoef Exp $
 */


/* include files for the application */
#include "custom.h"

/* define's here */

/* extra variables */
long permdegree;        /* degree of permutation group G */
void *minptpart;        /* pointer to a point used in DC_FINDPART */
 
/* Some of the declarations of functions below may be  substituted by 
   macros for some applications */


/* void DC_CLIENT_INIT(FILE *input) 
   this function must set at least the following global variables for the
   client, see the beginning of `dcclientgeneric.c':
   
long dc_length_pt;      length of point in orbit in bytes   
long dc_ng;             number of elements to condense (in startup number
                        of given generators of G   
void **ggen;            list of dc_ng void pointers to generators of G   
long dc_nk;             number of given generators of K   
void **kgen;            list of dc_nk void pointers to generators of K   
void *start;            pointer to start point
long htklen;            length of hash table for K-orbits   
long maxminpart;        maximal number of minimal elements in a part the 
			partitions of a K-orbit
long maxpart;           maximal number of elements in a part
long maxnrpartskorb;    maximal number of parts in a K-orbit
long maxextra;          maximal extra number of minimal representatives per
			K-orbit 
long dc_max_fact_req;   this times nr of hash servers requests to hash 
                        servers are produced at most in `CondMatRows'   

Furthermore it should set up all global data needed by the other functions
defined below.

*/

void DC_CLIENT_INIT(FILE *input)
{
  long i, j;
  int *ptr;
  char *cptr;

  /* round to multiple of 4 + sizeof(long)
     since operation is so cheap here we will compute the actual hash values 
     also in the client and put it in front of the data, then the
     hash function in the server will become *very* cheap. */
  fscanf(input,"%ld",&dc_length_pt);
  minptpart = (void*)malloc(dc_length_pt);

  /* the degree of the permutations */
  fscanf(input, "%ld", &permdegree);
  /* number of elements to condense */
  fscanf(input, "%ld", &dc_ng);
  /* elements of G to condense */
  ggen = (void*)malloc(dc_ng*sizeof(void*));
  for (i=0; i<dc_ng; i++) {
    ggen[i] = (void*)malloc(sizeof(int)*permdegree);
    for (j=0, ptr=(int*)ggen[i]; j<permdegree; ptr++, j++) 
      fscanf(input, "%d", ptr);
  }
  /* number of generators of K */
  fscanf(input, "%ld", &dc_nk);
  /* generators of K */
  kgen = (void*)malloc(dc_nk*sizeof(void*));
  for (i=0; i<dc_nk; i++) {
    kgen[i] = (void*)malloc(sizeof(int)*permdegree);
    for (j=0, ptr=(int*)kgen[i]; j<permdegree; ptr++, j++) 
      fscanf(input, "%d", ptr);
  }
  /* start point, here we assume action on list of bytes, initialize
     by zero to clear unused parts, beginning reserved for hash value */
  start = (void*)calloc(dc_length_pt, sizeof(char));
  for (i=0, cptr=((char*)start)+sizeof(long); i<permdegree; i++, cptr++) {
    fscanf(input, "%ld", &j);
    *cptr = j;
  }
  /* length of hash tables for K-orbits */
  fscanf(input, "%ld", &htklen);
  /* maximal number of minimal elements in part of the partitions of K-orbit */
  fscanf(input, "%ld", &maxminpart);
  /* maximal number of elements in a part */
  fscanf(input, "%ld", &maxpart);
  /* maximal number of parts in a K-orbit */
  fscanf(input, "%ld", &maxnrpartskorb);
  /* maximal extra number of minimal representatives per K-orbit */ 
  fscanf(input, "%ld", &maxextra);
  /* this times nr of hash servers requests to hash servers are 
     produced at most in `CondMatRows' */   
  fscanf(input, "%ld", &dc_max_fact_req);
}

/* void DC_OPERATION(void* pt, void *g, void *img)
   This should compute the image of point pt under group element g and 
   store the resulting point in img. Points and group elements are referred 
   to by void pointers, img must point to a large enough part of the memory 
   to hold the result.   
*/
void DC_OPERATION(void* pt, void *g, void *img)
{
  long i;
  int *gg;
  char *p, *pi;
  
  gg = (int*)g;
  p = ((char*)pt)+sizeof(long);
  pi = ((char*)img)+sizeof(long);
  for (i=0; i<permdegree; pi[gg[i]-1] = p[i], i++);
}

/* int DC_CMP_PTS(void* pt1, void* pt2, size_t len) 
   This is for comparing two points (of data length `len'), returns 
   0 in case of equality. 

   done by ???????????????????????
   #define DC_CMP_PTS memcmp
*/
int DC_CMP_PTS(void *pt1, void *pt2, size_t len)
{
  return memcmp(pt1, pt2, len);
}

/* long DC_HASHFUN(void *pt, size_t len);
   The hash function, it gets a point as one argument and the length of 
   its data, it has to return a long number.
*/
long DC_HASHFUN(void *pt, size_t len)
{
  len = len; /* unused */
  /* very easy here, since the hashvalue is the beginning of the data */
  return *((long*)pt);
}

/* long DC_HASHSERVFUN(void *);
   The hash server function, it gets a point as one argument and has
   to return the number of a hash server in the range from
   0..#hashservers-1: */
long DC_HASHSERVFUN(void *pt)
{ 
  pt = pt; /* unused */
  /* just one hash server */
  if (nrhashserv == 1) return 0;
  else return (*((unsigned long *)pt) / 100) % nrhashserv;
}

/* void DC_FINDMINIMAL(void *pt1, void *pt2);
   This function is called with a point pt1 and has to store one 
   minimal point in the same part of the K-orbit into pt2. 
   pt2 must point to a large enough part of the memory.
*/
/* the *actual* hash function, puts hash value into beginning of pt */
void actualhash(void * pt)
{
  long i, res=0;
  char *p = ((char*)pt)+sizeof(long);
  for (i=0; i<permdegree; i++)
    res = res*137+p[i];
  *((long*)pt) = res;
}

/* -------------    see another one below ----------------- */
/* this is *very* specific to one application,
     K subset S_21 contains subgroup 3^7 = <(1,2,3),(4,5,6), ...> */  
/*void DC_FINDMINIMAL(void *pt1, void *pt2)
{ 
  long i;
  char *p1, *p2;
  p1 = (char*)pt1+sizeof(long);
  p2 = (char*)pt2+sizeof(long);
  for (i=0; i<21; i+=3, p1+=3, p2+=3) {
    if (p1[0]<=p1[1]) {
      if (p1[0]<p1[2]) {
	p2[0]=p1[0]; p2[1]=p1[1]; p2[2]=p1[2];
      }
      else {
	p2[0]=p1[2]; p2[1]=p1[0]; p2[2]=p1[1];	
      }
    }
    else {
      if (p1[1]<=p1[2]) {
	p2[0]=p1[1]; p2[1]=p1[2]; p2[2]=p1[0];	
      }
      else {
	p2[0]=p1[2]; p2[1]=p1[0]; p2[2]=p1[1];	
      }
    }
  }
for minimals we always compute the hash value 
  actualhash(pt2);
}*/
  
/* this is *very* specific to one application,
   K subset S_21 contains subgroup S_3^7 = <(1,2,3),(1,2),(4,5,6), ...> */  
void DC_FINDMINIMAL(void *pt1, void *pt2)
{ 
  long i;
  char *p1, *p2;
  p1 = (char*)pt1+sizeof(long);
  p2 = (char*)pt2+sizeof(long);
  /* sort three entries */
  for (i=0; i<21; i+=3, p1+=3, p2+=3) {
    if (p1[1]<p1[0]) {
      if (p1[2]<p1[1]) {
	p2[0] = p1[2]; p2[1] = p1[1]; p2[2] = p1[0];
      }
      else {
	if (p1[2]<p1[0]) {
	  p2[0] = p1[1]; p2[1] = p1[2]; p2[2] = p1[0];
	}
	else {
	  p2[0] = p1[1]; p2[1] = p1[0]; p2[2] = p1[2];
	}
      }
    }
    else {
      if (p1[2]<p1[0]) {
	p2[0] = p1[2]; p2[1] = p1[0]; p2[2] = p1[1];
      }
      else {
	if (p1[2]<p1[1]) {
	  p2[0] = p1[0]; p2[1] = p1[2]; p2[2] = p1[1];
	}
	else {
	  p2[0] = p1[0]; p2[1] = p1[1]; p2[2] = p1[2];
	}
      }
    }
  }

  /* for minimals we always compute the hash value */
  actualhash(pt2);
}

/* void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   minimal points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found minimal points.
*/
void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr)
{
  DC_FINDMINIMAL(pt1, ptr);
  *nr = 1;
}

/* void DC_FINDPART(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found  points.
*/
/* this is also very special for one application, part is
   suborbit of S_3^7 */
void DC_FINDPART(void *pt1, void *ptr, long *nr)
{
  long i, n, j, r;
  char max[7], *cp, *cpm;

  /* corresponding minimal */
  DC_FINDMINIMAL(pt1, minptpart);
  /* size of orbit of S_3 on 3-tuple */
  for (i=0, cp=((char*)minptpart)+sizeof(long); i<7; cp+=3, i++) {
    if (cp[0]==cp[2]) 
      max[i] = 1;
    else {
      if (cp[0]==cp[1]) 
	max[i] = 3;
      else if (cp[1]==cp[2]) {
	/* change abb to bba */
	cp[2] = cp[0];
	cp[0] = cp[1];
	max[i] = 3;
      }
      else
	max[i] = 6;
    }
  }
  /* length of part */
  for (i=0, *nr=1; i<7; ((*nr)*=(max[i])), i++);
  /* evaluate part */
  for (n=0; n<(*nr); n++) {
    r = n;
    cp = ((char*)ptr)+(n*dc_length_pt+sizeof(long));
    cpm=((char*)minptpart)+sizeof(long);
    for (i=0; i<7; cp+=3, cpm+=3, i++) {
      j = r % max[i];
      r = (r-j)/max[i];
      if (max[i]==1) {
	cp[0] = cpm[0]; cp[1] = cpm[0]; cp[2] = cpm[0];
      }
      else {
	if (max[i]==3) {
	  switch (j) {
	  case 0: {cp[0] = cpm[0]; cp[1] = cpm[0]; cp[2] = cpm[2];break;}
	  case 1: {cp[0] = cpm[0]; cp[1] = cpm[2]; cp[2] = cpm[0];break;}
	  default: {cp[0] = cpm[2]; cp[1] = cpm[0]; cp[2] = cpm[0];break;}
	  }
	}
	else {
	  switch(j) {
	  case 0: {cp[0] = cpm[0]; cp[1] = cpm[1]; cp[2] = cpm[2];break;}
	  case 1: {cp[0] = cpm[1]; cp[1] = cpm[0]; cp[2] = cpm[2];break;}
	  case 2: {cp[0] = cpm[0]; cp[1] = cpm[2]; cp[2] = cpm[1];break;}
	  case 3: {cp[0] = cpm[2]; cp[1] = cpm[1]; cp[2] = cpm[0];break;}
	  case 4: {cp[0] = cpm[1]; cp[1] = cpm[2]; cp[2] = cpm[0];break;}
	  default: {cp[0] = cpm[2]; cp[1] = cpm[0]; cp[2] = cpm[1];break;}
	  }
	}
      }
    }
  }
}


